# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# classes.py
# Copyright (C) 2014-2021 Fracpete (pythonwekawrapper at gmail dot com)

import os
import inspect
import argparse
import json
import logging
import re
import csv
import weka.core.typeconv as typeconv
import javabridge
from javabridge import JWrapper, JClassWrapper
from javabridge.jutil import JavaException
import weka.core.jvm as jvm

# logging setup
logger = logging.getLogger("weka.core.classes")

suggestions = None
""" dictionary for class -> package relation """


def get_class(classname):
    """
    Returns the class object associated with the dot-notation classname.

    Taken from here: http://stackoverflow.com/a/452981

    :param classname: the classname
    :type classname: str
    :return: the class object
    :rtype: object
    """
    parts = classname.split('.')
    module = ".".join(parts[:-1])
    m = __import__(module)
    for comp in parts[1:]:
        m = getattr(m, comp)
    return m


def get_jclass(classname):
    """
    Returns the Java class object associated with the dot-notation classname.

    :param classname: the classname
    :type classname: str
    :return: the class object
    :rtype: JB_Object
    """
    try:
        return javabridge.class_for_name(classname)
    except:
        try:
            return javabridge.static_call(
                "Lweka/core/ClassHelper;", "forName",
                "(Ljava/lang/Class;Ljava/lang/String;)Ljava/lang/Class;",
                javabridge.class_for_name("java.lang.Object"), classname)
        except:
            return javabridge.static_call(
                "Lweka/core/ClassHelper;", "newInstance",
                "(Ljava/lang/String;[Ljava/lang/Class;[Ljava/lang/Object;)Ljava/lang/Object;",
                classname, None, None)


def get_enum(classname, enm):
    """
    Returns the instance of the enum.

    :param classname: the classname of the enum
    :type classname: str
    :param enm: the name of the enum element to return
    :type enm: str
    :return: the enum instance
    :rtype: JB_Object
    """
    return javabridge.static_call(
        "Lweka/core/ClassHelper;", "getEnum",
        "(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/Object;",
        classname, enm)


def get_static_field(classname, fieldname, signature):
    """
    Returns the Java object associated with the static field of the specified class.

    :param classname: the classname of the class to get the field from
    :type classname: str
    :param fieldname: the name of the field to retriev
    :type fieldname: str
    :return: the object
    :rtype: JB_Object
    """
    try:
        return javabridge.get_static_field(classname, fieldname, signature)
    except:
        return javabridge.static_call(
            "Lweka/core/ClassHelper;", "getStaticField",
            "(Ljava/lang/String;Ljava/lang/String;)Ljava/lang/Object;",
            classname, fieldname)


def get_classname(obj):
    """
    Returns the classname of the JB_Object, Python class or object.

    :param obj: the java object or Python class/object to get the classname for
    :type obj: object
    :return: the classname
    :rtype: str
    """
    if isinstance(obj, javabridge.JB_Object):
        cls = javabridge.call(obj, "getClass", "()Ljava/lang/Class;")
        return javabridge.call(cls, "getName", "()Ljava/lang/String;")
    elif inspect.isclass(obj):
        return obj.__module__ + "." + obj.__name__
    else:
        return get_classname(obj.__class__)


def is_instance_of(obj, class_or_intf_name):
    """
    Checks whether the Java object implements the specified interface or is a subclass of the superclass.

    :param obj: the Java object to check
    :type obj: JB_Object
    :param class_or_intf_name: the superclass or interface to check, dot notation or with forward slashes
    :type class_or_intf_name: str
    :return: true if either implements interface or subclass of superclass
    :rtype: bool
    """
    class_or_intf_name = class_or_intf_name.replace("/", ".")
    classname = get_classname(obj)
    # array? retrieve component type and check that
    if is_array(obj):
        jarray = JavaArray(jobject=obj)
        classname = jarray.component_type()
    result = javabridge.static_call(
        "Lweka/core/InheritanceUtils;", "isSubclass",
        "(Ljava/lang/String;Ljava/lang/String;)Z",
        class_or_intf_name, classname)
    if result:
        return True
    return javabridge.static_call(
        "Lweka/core/InheritanceUtils;", "hasInterface",
        "(Ljava/lang/String;Ljava/lang/String;)Z",
        class_or_intf_name, classname)


def is_array(obj):
    """
    Checks whether the Java object is an array.

    :param obj: the Java object to check
    :type obj: JB_Object
    :return: whether the object is an array
    :rtype: bool
    """
    cls = javabridge.call(obj, "getClass", "()Ljava/lang/Class;")
    return javabridge.call(cls, "isArray", "()Z")


def list_property_names(obj):
    """
    Lists the property names (Bean properties, ie read/write method pair) of the Java object.

    :param obj: the object to inspect
    :type obj: JB_Object or JavaObject
    :return: the list of property names
    :rtype: list
    """
    if isinstance(obj, JavaObject):
        obj = obj.jobject

    return typeconv.jstring_array_to_list(
        javabridge.static_call(
            "Lweka/core/ClassHelper;", "listPropertyNames",
            "(Ljava/lang/Object;)[Ljava/lang/String;",
            obj))


def new_instance(classname):
    """
    Instantiates an object of the specified class.

    :param classname: the name of the class to instantiate
    :type classname: str
    :return: the object
    :rtype: JB_Object
    """
    return javabridge.static_call(
        "Lweka/core/ClassHelper;", "newInstance",
        "(Ljava/lang/String;[Ljava/lang/Class;[Ljava/lang/Object;)Ljava/lang/Object;",
        classname, None, None)

class Stoppable(object):
    """
    Classes that can be stopped.
    """

    def is_stopped(self):
        """
        Returns whether the object has been stopped.

        :return: whether stopped
        :rtype: bool
        """
        raise Exception("Not implemented!")

    def stop_execution(self):
        """
        Triggers the stopping of the object.
        """
        raise Exception("Not implemented!")


from_dict_handlers = {}
"""
The methods that handle the restoration from a JSON dictionary, stored under their 'type'.
"""


def register_dict_handler(typestr, handler):
    """
    Registers a handler for restoring an object from a JSON dictionary.

    :param typestr: the type of the object
    :type typestr: str
    :param handler: the method
    """
    global from_dict_handlers
    from_dict_handlers[typestr] = handler


def deregister_dict_handler(typestr):
    """
    Deregisters a handler for restoring an object from a JSON dictionary.

    :param typestr: the type of the object
    :type typestr: str
    """
    global from_dict_handlers
    del from_dict_handlers[typestr]


def has_dict_handler(typestr):
    """
    Returns the handler for restoring an object from a JSON dictionary.

    :param typestr: the type of the object
    :type typestr: str
    :return: the handler, None if not available
    """
    global from_dict_handlers
    return typestr in from_dict_handlers


def get_dict_handler(typestr):
    """
    Returns the handler for restoring an object from a JSON dictionary.

    :param typestr: the type of the object
    :type typestr: str
    :return: the handler, None if not available
    """
    global from_dict_handlers
    return from_dict_handlers[str(typestr)]


class JSONObject(object):
    """
    Ancestor for classes that can be represented as JSON and restored from JSON.
    """

    def to_dict(self):
        """
        Returns a dictionary that represents this object, to be used for JSONification.

        :return: the object dictionary
        :rtype: dict
        """
        raise Exception("Not implemented!")

    @classmethod
    def from_dict(cls, d):
        """
        Restores an object state from a dictionary, used in de-JSONification.

        :param d: the object dictionary
        :type d: dict
        :return: the object
        :rtype: object
        """
        raise Exception("Not implemented!")

    def to_json(self):
        """
        Returns the options as JSON.

        :return: the object as string
        :rtype: str
        """
        return json.dumps(self.to_dict(), sort_keys=True, indent=2, separators=(',', ': '))

    @classmethod
    def from_json(cls, s):
        """
        Restores the object from the given JSON.

        :param s: the JSON string to parse
        :type s: str
        :return: the
        """
        d = json.loads(s)
        return get_dict_handler(d["type"])(d)

    def shallow_copy(self):
        """
        Returns a shallow copy of itself.

        :return: the copy
        :rtype: object
        """
        return self.from_json(self.to_json())


class Configurable(JSONObject):
    """
    The ancestor for all actors.
    """

    def __init__(self, config=None):
        """
        Initializes the object.

        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        self._logger = None
        self._help = {}
        self._config = self.fix_config({})
        if config is not None:
            self.config = config
        if not has_dict_handler("Configurable"):
            register_dict_handler("Configurable", Configurable.from_dict)

    def __repr__(self):
        """
        Returns Python code for instantiating the object.

        :return: the representation
        :rtype: str
        """
        return \
            self.__class__.__module__ + "." + self.__class__.__name__ \
            + "(config=" + str(self.config) + ")"

    def description(self):
        """
        Returns a description of the object.

        :return: the description
        :rtype: str
        """
        raise Exception("Not implemented!")

    def fix_config(self, options):
        """
        Fixes the options, if necessary. I.e., it adds all required elements to the dictionary.

        :param options: the options to fix
        :type options: dict
        :return: the (potentially) fixed options
        :rtype: dict
        """
        return options

    @property
    def config(self):
        """
        Obtains the currently set options of the actor.

        :return: the options
        :rtype: dict
        """
        return self._config

    @config.setter
    def config(self, options):
        """
        Sets the options of the actor.

        :param options: the options
        :type options: dict
        """
        self._config = self.fix_config(options)

    def to_dict(self):
        """
        Returns a dictionary that represents this object, to be used for JSONification.

        :return: the object dictionary
        :rtype: dict
        """
        result = {}
        result["type"] = "Configurable"
        result["class"] = get_classname(self)
        result["config"] = {}
        for k in self._config:
            v = self._config[k]
            if isinstance(v, JSONObject):
                result["config"][k] = v.to_dict()
            else:
                result["config"][k] = v
        return result

    @classmethod
    def from_dict(cls, d):
        """
        Restores its state from a dictionary, used in de-JSONification.

        :param d: the object dictionary
        :type d: dict
        """
        conf = {}
        for k in d["config"]:
            v = d["config"][k]
            if isinstance(v, dict):
                conf[str(k)] = get_dict_handler(d["config"]["type"])(v)
            else:
                conf[str(k)] = v
        return get_class(str(d["class"]))(config=conf)

    def new_logger(self):
        """
        Returns a new logger instance.

        :return: the logger instance
        :rtype: logger
        """
        return logging.getLogger(get_classname(self))

    @property
    def logger(self):
        """
        Returns the logger object.

        :return: the logger
        :rtype: logger
        """
        if self._logger is None:
            self._logger = self.new_logger()
        return self._logger

    @property
    def help(self):
        """
        Obtains the help information per option for this actor.

        :return: the help
        :rtype: dict
        """
        return self._help

    def generate_help(self):
        """
        Generates a help string for this actor.

        :return: the help string
        :rtype: str
        """
        result = []
        result.append(self.__class__.__name__)
        result.append(re.sub(r'.', '=', self.__class__.__name__))
        result.append("")
        result.append("DESCRIPTION")
        result.append(self.description())
        result.append("")
        result.append("OPTIONS")
        opts = sorted(self.config.keys())
        for opt in opts:
            result.append(opt)
            helpstr = self.help[opt]
            if helpstr is None:
                helpstr = "-missing help-"
            result.append("\t" + helpstr)
            result.append("")
        return '\n'.join(result)

    def print_help(self):
        """
        Prints a help string for this actor to stdout.
        """
        print(self.generate_help())


class JavaObject(JSONObject):
    """
    Basic Java object.
    """
    
    def __init__(self, jobject):
        """
        Initializes the wrapper with the specified Java object.

        :param jobject: the Java object to wrap
        :type jobject: JB_Object
        """
        if jobject is None:
            raise Exception("No Java object supplied!")
        self.jobject = jobject
        if not has_dict_handler("JavaObject"):
            register_dict_handler("JavaObject", JavaObject.from_dict)

    def __str__(self):
        """
        Just calls the toString() method.

        :rtype: str
        """
        return javabridge.to_string(self.jobject)

    def __repr__(self):
        """
        Just calls the toString() method.

        :rtype: str
        """
        return javabridge.to_string(self.jobject)

    def __unicode__(self):
        """
        Just calls the toString() method.

        :rtype: str
        """
        return javabridge.to_string(self.jobject)

    @property
    def classname(self):
        """
        Returns the Java classname in dot-notation.

        :return: the Java classname
        :rtype: str
        """
        cls = javabridge.call(self.jobject, "getClass", "()Ljava/lang/Class;")
        return javabridge.call(cls, "getName", "()Ljava/lang/String;")

    def set_property(self, path, jobject):
        """
        Attempts to set the value (jobject, a Java object) of the provided (bean) property path.

        :param path: the property path, e.g., "filter" for a setFilter(...)/getFilter() method pair
        :type path: str
        :param jobject: the Java object to set; if instance of JavaObject class, the jobject member is automatically used
        :type jobject: JB_Object
        """
        # unwrap?
        if isinstance(jobject, JavaObject):
            jobject = jobject.jobject

        javabridge.static_call(
            "Lweka/core/PropertyPath;", "setValue",
            "(Ljava/lang/Object;Ljava/lang/String;Ljava/lang/Object;)V",
            self.jobject, path, jobject)

    def get_property(self, path):
        """
        Attempts to get the value (jobject, a Java object) of the provided (bean) property path.

        :param path: the property path, e.g., "filter" for a setFilter(...)/getFilter() method pair
        :type path: str
        :return: the wrapped Java object
        :rtype: JavaObject
        """
        return JavaObject(javabridge.static_call(
            "Lweka/core/PropertyPath;", "getValue",
            "(Ljava/lang/Object;Ljava/lang/String;)Ljava/lang/Object;",
            self.jobject, path))

    @property
    def jclass(self):
        """
        Returns the Java class object of the underlying Java object.

        :return: the Java class
        :rtype: JB_Object
        """
        return javabridge.call(self.jobject, "getClass", "()Ljava/lang/Class;")

    @property
    def jwrapper(self):
        """
        Returns a JWrapper instance of the encapsulated Java object, giving access to methods
        using dot notation.

        http://pythonhosted.org//javabridge/highlevel.html#wrapping-java-objects-using-reflection

        :return: the wrapper
        :rtype: JWrapper
        """
        return JWrapper(self.jobject)

    @property
    def jclasswrapper(self):
        """
        Returns a JClassWrapper instance of the class for the encapsulated Java object, giving
        access to the class methods using dot notation.

        http://pythonhosted.org//javabridge/highlevel.html#wrapping-java-objects-using-reflection

        :return: the wrapper
        :rtype: JClassWrapper
        """
        return JClassWrapper(javabridge.call(self.jobject, "getClass", "()Ljava/lang/Class;"))

    @property
    def is_serializable(self):
        """
        Returns true if the object is serialiable.

        :return: true if serializable
        :rtype: bool
        """
        return JavaObject.check_type(self.jobject, "java.io.Serializable")

    def to_dict(self):
        """
        Returns a dictionary that represents this object, to be used for JSONification.

        :return: the object dictionary
        :rtype: dict
        """
        result = {}
        result["type"] = "JavaObject"
        result["class"] = self.classname
        return result

    @classmethod
    def from_dict(cls, d):
        """
        Restores an object state from a dictionary, used in de-JSONification.

        :param d: the object dictionary
        :type d: dict
        :return: the object
        :rtype: object
        """
        return JavaObject(cls.new_instance(d["class"]))

    @classmethod
    def check_type(cls, jobject, intf_or_class):
        """
        Returns whether the object implements the specified interface or is a subclass.

        :param jobject: the Java object to check
        :type jobject: JB_Object
        :param intf_or_class: the classname in Java notation (eg "weka.core.DenseInstance;")
        :type intf_or_class: str
        :return: whether object implements interface or is subclass
        :rtype: bool
        """
        return is_instance_of(jobject, intf_or_class)
        
    @classmethod
    def enforce_type(cls, jobject, intf_or_class):
        """
        Raises an exception if the object does not implement the specified interface or is not a subclass. 

        :param jobject: the Java object to check
        :type jobject: JB_Object
        :param intf_or_class: the classname in Java notation (eg "weka.core.DenseInstance")
        :type intf_or_class: str
        """
        if not cls.check_type(jobject, intf_or_class):
            raise TypeError("Object does not implement or subclass " + intf_or_class + ": " + get_classname(jobject))

    @classmethod
    def new_instance(cls, classname, options=None):
        """
        Creates a new object from the given classname using the default constructor, None in case of error.

        :param classname: the classname in Java notation (eg "weka.core.DenseInstance")
        :type classname: str
        :param options: the list of options to use, ignored if None
        :type options: list
        :return: the Java object
        :rtype: JB_Object
        """
        try:
            if options is None:
                options = []
            return javabridge.static_call(
                "Lweka/core/Utils;", "forName",
                "(Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/String;)Ljava/lang/Object;",
                javabridge.class_for_name("java.lang.Object"), classname, options)
        except JavaException as e:
            print("Failed to instantiate " + classname + ": " + str(e))
            suggestions = suggest_package(classname, exact=True)
            if len(suggestions) > 0:
                print("Class '" + classname + "' is available from package: " + ", ".join(suggestions))
                if not jvm.with_package_support:
                    print("Also, you need to start up the JVM with package support:\njvm.start(packages=True)")
            else:
                if not jvm.with_package_support:
                    print("If this class is from a Weka package, then you need to start up the JVM with package support:\njvm.start(packages=True)")
            return None


class Environment(JavaObject):
    """
    Wraps around weka.core.Environment
    """

    def __init__(self, jobject=None):
        """
        Initializes the object.

        :param jobject: the environment java object to use
        :type jobject: JB_Object
        """
        if jobject is None:
            jobject = javabridge.make_instance("Lweka/core/Environment;", "()V")
        else:
            Environment.enforce_type(jobject, "weka.core.Environment")
        super(Environment, self).__init__(jobject)

    def add_variable(self, key, value, system_wide=False):
        """
        Adds the environment variable.

        :param key: the name of the variable
        :type key: str
        :param value: the value
        :type value: str
        :param system_wide: whether to add the variable system wide
        :type system_wide: bool
        """
        if system_wide:
            javabridge.call(self.jobject, "addVariableSystemWide", "(Ljava/lang/String;Ljava/lang/String;)V", key, value)
        else:
            javabridge.call(self.jobject, "addVariable", "(Ljava/lang/String;Ljava/lang/String;)V", key, value)

    def remove_variable(self, key):
        """
        Adds the environment variable.

        :param key: the name of the variable
        :type key: str
        """
        javabridge.call(self.jobject, "removeVariable", "(Ljava/lang/String;)V", key)

    def variable_value(self, key):
        """
        Returns the value of the environment variable.

        :param key: the name of the variable
        :type key: str
        :return: the variable value
        :rtype: str
        """
        return javabridge.call(self.jobject, "getVariableValue", "(Ljava/lang/String;)Ljava/lang/String;", key)

    def variable_names(self):
        """
        Returns the names of all environment variables.

        :return: the names of the variables
        :rtype: list
        """
        result = []
        names = javabridge.call(self.jobject, "getVariableNames", "()Ljava/util/Set;")
        for name in javabridge.iterate_collection(names):
            result.append(javabridge.to_string(name))
        return result

    @classmethod
    def system_wide(cls):
        """
        Returns the system-wide environment.

        ;return: the environment
        :rtype: Environment
        """
        return Environment(jobject=javabridge.static_call(
            "Lweka/core/Environment;", "getSystemWide", "()Lweka/core/Environment;"))


class JavaArrayIterator(object):
    """
    Iterator for elements in a Java array.
    """
    def __init__(self, data):
        """
        :param data: the Java array to iterate over
        :type data: JavaArray
        """
        self.data = data
        self.index = 0
        self.length = len(data)

    def __iter__(self):
        """
        Returns itself.
        """
        return self

    def __next__(self):
        """
        Returns the next element from the array.

        :return: the next array element object, wrapped as JavaObject if not null
        :rtype: JavaObject or None
        """
        if self.index < self.length:
            index = self.index
            self.index += 1
            return self.data[index]
        else:
            raise StopIteration()


class JavaArray(JavaObject):
    """
    Convenience wrapper around Java arrays.
    """

    def __init__(self, jobject):
        """
        Initializes the wrapper with the specified Java object.

        :param jobject: the java array object to wrap
        :type jobject: JB_Object
        """
        super(JavaArray, self).__init__(jobject)
        c = self.jclass
        if not javabridge.call(c, "isArray", "()Z"):
            raise Exception("Not an array!")

    def __len__(self):
        """
        Returns the length of the array.

        :return: the array length
        :rtype: int
        """
        return javabridge.get_env().get_array_length(self.jobject)

    def __getitem__(self, key):
        """
        Returns the specified element in the array wrapped in a JavaObject.

        :param key: the index of the element to retrieve
        :type key: int
        :return: the element or None if element is null
        :rtype; JavaObject
        """
        if not isinstance(key, int):
            raise Exception("Key must be an integer!")
        element = javabridge.static_call(
            "Ljava/lang/reflect/Array;", "get", "(Ljava/lang/Object;I)Ljava/lang/Object;",
            self.jobject, key)
        if element is None:
            return None
        return JavaObject(element)

    def __setitem__(self, key, value):
        """
        Sets the specified element in the array.

        :param key: the index of the element to set
        :type key: int
        :param value: the object to set (JavaObject or JB_Object)
        """
        if isinstance(value, JavaObject):
            obj = value.jobject
        else:
            obj = value
        if not isinstance(key, int):
            raise Exception("Key must be an integer!")
        javabridge.static_call(
            "Ljava/lang/reflect/Array;", "set", "(Ljava/lang/Object;ILjava/lang/Object;)V",
            self.jobject, key, obj)

    def __delitem__(self, key):
        """
        Not implemented, raises an exception.
        """
        raise Exception("Cannot delete item from array!")

    def __iter__(self):
        """
        Returns an iterator over the elements.

        :return: the iterator
        :rtype: JavaArrayIterator
        """
        return JavaArrayIterator(self)

    def component_type(self):
        """
        Returns the classname of the elements.

        :return: the class of the elements
        :rtype: str
        """
        cls = javabridge.call(self.jobject, "getClass", "()Ljava/lang/Class;")
        comptype = javabridge.call(cls, "getComponentType", "()Ljava/lang/Class;")
        return javabridge.call(comptype, "getName", "()Ljava/lang/String;")

    @classmethod
    def new_instance(cls, classname, length):
        """
        Creates a new array with the given classname and length; initial values are null.

        :param classname: the classname in Java notation (eg "weka.core.DenseInstance")
        :type classname: str
        :param length: the length of the array
        :type length: int
        :return: the Java array
        :rtype: JB_Object
        """
        return javabridge.static_call(
            "Ljava/lang/reflect/Array;",
            "newInstance",
            "(Ljava/lang/Class;I)Ljava/lang/Object;",
            get_jclass(classname=classname), length)


class Enum(JavaObject):
    """
    Wrapper for Java enums.
    """

    def __init__(self, jobject=None, enum=None, member=None):
        """
        Initializes the enum class.

        :param jobject: the Java object to wrap
        :type jobject: JB_Object
        :param enum: the enum class to instantiate (dot notation)
        :type enum: str
        :param member: the member of the enum class to instantiate
        :type member: str
        """
        if jobject is None:
            if enum is None:
                raise Exception("Enum member cannot be None if no JB_Object instance provided!")
            jobject = get_enum(enum, member)
        super(Enum, self).__init__(jobject)

    @property
    def name(self):
        """
        Returns the name of the enum member.

        :return: the name
        :rtype: str
        """
        return javabridge.call(self.jobject, "name", "()Ljava/lang/String;")

    @property
    def ordinal(self):
        """
        Returns the ordinal of the enum member.

        :return: the ordinal
        :rtype: int
        """
        return javabridge.call(self.jobject, "ordinal", "()I")

    @property
    def values(self):
        """
        Returns list of all enum members.

        :return: all enum members
        :rtype: list
        """
        cls = javabridge.call(self.jobject, "getClass", "()Ljava/lang/Class;")
        clsname = javabridge.call(cls, "getName", "()Ljava/lang/String;")
        l = javabridge.static_call(clsname.replace(".", "/"), "values", "()[L" + clsname.replace(".", "/") + ";")
        l = javabridge.get_env().get_object_array_elements(l)
        result = []
        for item in l:
            result.append(Enum(jobject=item))
        return result


class Random(JavaObject):
    """
    Wrapper for the java.util.Random class.
    """

    def __init__(self, seed):
        """
        The seed value.

        :param seed: the seed value
        :type seed: int
        """
        super(Random, self).__init__(javabridge.make_instance("Ljava/util/Random;", "(J)V", seed))

    def next_int(self, n=None):
        """
        Next random integer. if n is provided, then between 0 and n-1.

        :param n: the upper limit (minus 1) for the random integer
        :type n: int
        :return: the next random integer
        :rtype: int
        """
        if n is None:
            return javabridge.call(self.jobject, "nextInt", "()I")
        else:
            return javabridge.call(self.jobject, "nextInt", "(I)I", n)

    def next_double(self):
        """
        Next random double.

        :return: the next random double
        :rtype: double
        """
        return javabridge.call(self.jobject, "nextDouble", "()D")


class Option(JavaObject):
    """
    Wrapper for the weka.core.Option class.
    """

    def __init__(self, jobject):
        """
        Initializes the wrapper with the specified option object.

        :param jobject: the java object to wrap
        :type jobject: JB_Object
        """
        super(Option, self).__init__(jobject)
        Option.enforce_type(jobject, "weka.core.Option")

    @property
    def name(self):
        """
        Returns the name of the option.

        :return: the name
        :rtype: str
        """
        return javabridge.call(self.jobject, "name", "()Ljava/lang/String;")

    @property
    def description(self):
        """
        Returns the description of the option.

        :return: the description
        :rtype: str
        """
        return javabridge.call(self.jobject, "description", "()Ljava/lang/String;")

    @property
    def synopsis(self):
        """
        Returns the synopsis of the option.

        :return: the synopsis
        :rtype: str
        """
        return javabridge.call(self.jobject, "synopsis", "()Ljava/lang/String;")

    @property
    def num_arguments(self):
        """
        Returns the synopsis of the option.

        :return: the synopsis
        :rtype: str
        """
        return javabridge.call(self.jobject, "numArguments", "()I")


class OptionHandler(JavaObject, Configurable):
    """
    Ancestor for option-handling classes. 
    Classes should implement the weka.core.OptionHandler interface to have any effect.
    """
    
    def __init__(self, jobject, options=None):
        """
        Initializes the wrapper with the specified Java object.

        :param jobject: the java object to wrap
        :type jobject: JB_Object
        :param options: the options to set
        :type options: list
        """
        super(OptionHandler, self).__init__(jobject)
        self.is_optionhandler = OptionHandler.check_type(jobject, "weka.core.OptionHandler")
        if (options is not None) and (len(options) > 0):
            self.options = options
        # we have to manually instantiate some objects, since multiple inheritance doesn't call
        # Configurable constructor
        self._logger = None
        self._help = {}
        self._config = self.fix_config({})
        if not has_dict_handler("OptionHandler"):
            register_dict_handler("OptionHandler", OptionHandler.from_dict)

    def global_info(self):
        """
        Returns the globalInfo() result, None if not available.

        :rtypes: str
        """
        try:
            return javabridge.call(self.jobject, "globalInfo", "()Ljava/lang/String;")
        except JavaException:
            return None

    def description(self):
        """
        Returns a description of the object.

        :return: the description
        :rtype: str
        """
        return self.global_info()

    @property
    def options(self):
        """
        Obtains the currently set options as list.

        :return: the list of options
        :rtype: list
        """
        if self.is_optionhandler:
            return typeconv.jstring_array_to_list(javabridge.call(self.jobject, "getOptions", "()[Ljava/lang/String;"))
        else:
            return []

    @options.setter
    def options(self, options):
        """
        Sets the command-line options (as list).

        :param options: the list of command-line options to set
        :type options: list
        """
        if self.is_optionhandler:
            javabridge.call(self.jobject, "setOptions", "([Ljava/lang/String;)V", typeconv.string_list_to_jarray(options))

    def to_commandline(self):
        """
        Generates a commandline string from the JavaObject instance.

        :return: the commandline string
        :rtype: str
        """
        return javabridge.static_call(
            "Lweka/core/Utils;", "toCommandLine",
            "(Ljava/lang/Object;)Ljava/lang/String;",
            self.jobject)

    def to_help(self):
        """
        Returns a string that contains the 'global_info' text and the options.

        :return: the generated help string
        :rtype: str
        """
        result = []
        result.append(self.classname)
        result.append("=" * len(self.classname))
        result.append("")
        result.append("DESCRIPTION")
        result.append("")
        result.append(self.global_info())
        result.append("")
        result.append("OPTIONS")
        result.append("")
        options = javabridge.call(self.jobject, "listOptions", "()Ljava/util/Enumeration;")
        enum = javabridge.get_enumeration_wrapper(options)
        while enum.hasMoreElements():
            opt = Option(enum.nextElement())
            result.append(opt.synopsis)
            result.append(opt.description)
            result.append("")
        return '\n'.join(result)

    def __str__(self):
        """
        Calls the toString() method of the java object.

        :return: the result of the toString() method
        :rtype: str
        """
        return javabridge.to_string(self.jobject)

    def to_dict(self):
        """
        Returns a dictionary that represents this object, to be used for JSONification.

        :return: the object dictionary
        :rtype: dict
        """
        result = super(OptionHandler, self).to_dict()
        result["type"] = "OptionHandler"
        result["options"] = join_options(self.options)
        return result

    @classmethod
    def from_dict(cls, d):
        """
        Restores an object state from a dictionary, used in de-JSONification.

        :param d: the object dictionary
        :type d: dict
        :return: the object
        :rtype: object
        """
        result = OptionHandler(cls.new_instance(d["class"]))
        result.options = split_options(d["options"])
        return result


class SingleIndex(JavaObject):
    """
    Wrapper for a Weka SingleIndex object.
    """

    def __init__(self, jobject=None, index=None):
        """
        Initializes the wrapper with the specified Java object or string index.

        :param jobject: the java object to wrap
        :type jobject: JB_Object
        :param index: the string index to use
        :type index: str
        """
        if jobject is None:
            if index is None:
                jobject = javabridge.make_instance("weka/core/SingleIndex", "()V")
            else:
                jobject = javabridge.make_instance("weka/core/SingleIndex", "(Ljava/lang/String;)V", index)
        else:
            self.enforce_type(jobject, "weka.core.SingleIndex")
        super(SingleIndex, self).__init__(jobject)

    def upper(self, upper):
        """
        Sets the upper limit.

        :param upper: the upper limit
        :type upper: int
        """
        javabridge.call(self.jobject, "setUpper", "(I)V", upper)

    def index(self):
        """
        Returns the integer index.

        :return: the 0-based integer index
        :rtype: int
        """
        return javabridge.call(self.jobject, "getIndex", "()I")

    @property
    def single_index(self):
        """
        Returns the string index.

        :return: the 1-based string index
        :rtype: str
        """
        return javabridge.call(self.jobject, "getSingleIndex", "()Ljava/lang/String;")

    @single_index.setter
    def single_index(self, index):
        """
        Sets the string index.

        :param index: the 1-based string index
        ::type index: str
        """
        javabridge.call(self.jobject, "setSingleIndex", "(Ljava/lang/String;)V", index)


class Range(JavaObject):
    """
    Wrapper for a Weka Range object.
    """

    def __init__(self, jobject=None, ranges=None):
        """
        Initializes the wrapper with the specified Java object or string range.

        :param jobject: the java object to wrap
        :type jobject: JB_Object
        :param ranges: the string range to use
        :type ranges: str
        """
        if jobject is None:
            if ranges is None:
                jobject = javabridge.make_instance("weka/core/Range", "()V")
            else:
                jobject = javabridge.make_instance("weka/core/Range", "(Ljava/lang/String;)V", ranges)
        else:
            self.enforce_type(jobject, "weka.core.Range")
        super(Range, self).__init__(jobject)

    def upper(self, upper):
        """
        Sets the upper limit.

        :param upper: the upper limit
        :type upper: int
        """
        javabridge.call(self.jobject, "setUpper", "(I)V", upper)

    def selection(self):
        """
        Returns the selection list.

        :return: the list of 0-based integer indices
        :rtype: list
        """
        return javabridge.get_env().get_int_array_elements(javabridge.call(self.jobject, "getSelection", "()[I"))

    @property
    def ranges(self):
        """
        Returns the string range.

        :return: the string range of 1-based indices
        :rtype: str
        """
        return javabridge.call(self.jobject, "getRanges", "()Ljava/lang/String;")

    @ranges.setter
    def ranges(self, rng):
        """
        Sets the string range.

        :param rng: the range to set
        :type rng: str
        """
        javabridge.call(self.jobject, "setRanges", "(Ljava/lang/String;)V", rng)

    @property
    def invert(self):
        """
        Returns whether the range is inverted.

        :return: true if inverted
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getInvert", "()Z")

    @invert.setter
    def invert(self, invert):
        """
        Sets the invert state.

        :param invert: whether to invert or not
        :type invert: bool
        """
        javabridge.call(self.jobject, "setInvert", "(Z)V", invert)


class Tag(JavaObject):
    """
    Wrapper for the weka.core.Tag class.
    """

    def __init__(self, jobject=None, ident=None, ident_str="", readable="", uppercase=True):
        """
        :param jobject: the java object to wrap
        :type jobject: JB_Object
        :param ident: the ID integer associated with the tag
        :type ident: int
        :param ident_str: the ID string associated with the tag (case-insensitive)
        :type ident_str: str
        :param readable: the text of the tag
        :type readable: str
        :param uppercase: whether to uppercase the id string
        :type uppercase: bool
        """
        if jobject is None:
            jobject = javabridge.make_instance(
                "weka/core/Tag", "(ILjava/lang/String;Ljava/lang/String;Z)V",
                ident, ident_str, readable, uppercase)
        else:
            self.enforce_type(jobject, "weka.core.Tag")
        super(Tag, self).__init__(jobject)

    @property
    def ident(self):
        """
        Returns the current integer ID of the tag.

        :return: the integer ID
        :rtype: int
        """
        return javabridge.call(self.jobject, "getID", "()I")

    @ident.setter
    def ident(self, value):
        """
        Sets the integer ID of the tag.

        :param value: the new ID
        :type value: int
        """
        javabridge.call(self.jobject, "setID", "(I)V", value)

    @property
    def identstr(self):
        """
        Returns the current ID string.

        :return: the ID string
        :rtype: str
        """
        return javabridge.call(self.jobject, "getIDStr", "()Ljava/lang/String;")

    @identstr.setter
    def identstr(self, value):
        """
        Sets the ID string.

        :param value: the new ID string
        :type value: str
        """
        javabridge.call(self.jobject, "setIDStr", "(Ljava/lang/String;)V", value)

    @property
    def readable(self):
        """
        Returns the 'human readable' string.

        :return: the readable string
        :rtype: str
        """
        return javabridge.call(self.jobject, "getReadable", "()Ljava/lang/String;")

    @readable.setter
    def readable(self, value):
        """
        Sets the 'human readable' string.

        :param value: the new readable string
        :type value: str
        """
        javabridge.call(self.jobject, "setReadable", "(Ljava/lang/String;)V", value)


class Tags(JavaObject):
    """
    Wrapper for an array of weka.core.Tag objects.
    """

    def __init__(self, jobject=None, tags=None):
        """
        :param jobject: the Java Tag array to wrap.
        :type jobject: JB_Object
        :param tags: the list of Tag objects to use
        :type tags: list
        """
        if tags is not None:
            jarray = JavaArray(JavaArray.new_instance("weka.core.Tag", len(tags)))
            for i in range(len(tags)):
                jarray[i] = tags[i]
            jobject = jarray.jobject
        self.enforce_type(jobject, "weka.core.Tag")
        super(Tags, self).__init__(jobject)
        self.array = JavaArray(self.jobject)

    def __len__(self):
        """
        Returns the number of Tag objects in the array.

        :return: the number of tag objects
        :rtype: int
        """
        return len(self.array)

    def __getitem__(self, item):
        """
        Returns the specified Tag from the array.

        :param item: the 0-based index
        :type item: int
        :return: the tag
        :rtype: Tag
        """
        return Tag(self.array[item].jobject)

    def __setitem__(self, key, value):
        """
        Not implemented.
        """
        raise Exception("Cannot set a Tag!")

    def __delitem__(self, key):
        """
        Not implemented.
        """
        raise Exception("Cannot delete a Tag!")

    def __str__(self):
        """
        Just calls the toString() method.

        :rtype: str
        """
        result = ""
        for i in range(len(self.array)):
            if i > 0:
                result += "|"
            result += str(self.array[i])
        return result

    def find(self, name):
        """
        Returns the Tag that matches the name.

        :param name: the string representation of the tag
        :type name: str
        :return: the tag, None if not found
        :rtype: Tag
        """
        result = None
        for t in self.array:
            if str(t) == name:
                result = Tag(t.jobject)
                break
        return result

    @classmethod
    def get_tags(cls, classname, field):
        """
        Instantiates the Tag array located in the specified class with the given field name.

        Example:
        tags = Tags.get_tags("weka.classifiers.functions.SMO", "TAGS_FILTER")

        :param classname: the classname in which the tags reside
        :type classname: str
        :param field: the field name of the Tag array
        :type field: str
        :return: the Tags objects
        :rtype: Tags
        """
        return Tags(jobject=get_static_field(classname, field, "[Lweka/core/Tag;"))

    @classmethod
    def get_object_tags(cls, javaobject, methodname):
        """
        Instantiates the Tag array obtained from the object using the specified method name.

        Example:
        cls = Classifier(classname="weka.classifiers.meta.MultiSearch")
        tags = Tags.get_object_tags(cls, "getMetricsTags")

        :param javaobject: the javaobject to obtain the tags from
        :type javaobject: JavaObject
        :param methodname: the method name returning the Tag array
        :type methodname: str
        :return: the Tags objects
        :rtype: Tags
        """
        return Tags(jobject=javabridge.call(javaobject.jobject, methodname, "()[Lweka/core/Tag;"))


class SelectedTag(JavaObject):
    """
    Wrapper for the weka.core.SelectedTag class.
    """

    def __init__(self, jobject=None, tag_id=None, tag_text=None, tags=None):
        """
        Initializes the wrapper with the specified Java object or tags and either tag_id or tag_text.

        :param jobject: the java object to wrap
        :type jobject: JB_Object
        :param tag_id: the integer associated with the tag
        :type tag_id: int
        :param tag_text: the text associated with the tag
        :type tag_text: str
        :param tags: list of Tag objects or Tags wrapper object
        :type tags: list or Tags
        """

        if isinstance(tags, Tags):
            tobj = tags.jobject
        else:
            tobj = tags

        if jobject is None:
            if tag_id is None:
                jobject = javabridge.make_instance(
                    "weka/core/SelectedTag", "(Ljava/lang/String;[Lweka/core/Tag;)V", tag_text, tobj)
            else:
                jobject = javabridge.make_instance(
                    "weka/core/SelectedTag", "(I[Lweka/core/Tag;)V", tag_id, tobj)
        else:
            self.enforce_type(jobject, "weka.core.SelectedTag")
        super(SelectedTag, self).__init__(jobject)

    @property
    def selected(self):
        """
        Returns the selected tag.

        :return: the tag
        :rtype: Tag
        """
        return Tag(javabridge.call(self.jobject, "getSelectedTag", "()Lweka/core/Tag;"))

    @property
    def tags(self):
        """
        Returns the associated tags.

        :return: the list of Tag objects
        :rtype: list
        """
        result = []
        a = javabridge.call(self.jobject, "getTags", "()Lweka/core/Tag;]")
        length = javabridge.get_env().get_array_length(a)
        wrapped = javabridge.get_env().get_object_array_elements(a)
        for i in range(length):
            result.append(Tag(javabridge.get_env().get_string(wrapped[i])))
        return result


def join_options(options):
    """
    Turns the list of options back into a single commandline string.

    :param options: the list of options to process
    :type options: list
    :return: the combined options
    :rtype: str
    """
    return javabridge.static_call(
        "Lweka/core/Utils;", "joinOptions",
        "([Ljava/lang/String;)Ljava/lang/String;",
        options)


def split_options(cmdline):
    """
    Splits the commandline into a list of options.

    :param cmdline: the commandline string to split into individual options
    :type cmdline: str
    :return: the split list of commandline options
    :rtype: list
    """
    return typeconv.jstring_array_to_list(
        javabridge.static_call(
            "Lweka/core/Utils;", "splitOptions",
            "(Ljava/lang/String;)[Ljava/lang/String;",
            cmdline))


def backquote(s):
    """
    Backquotes the string.

    :param s: the string to process
    :type s: str
    :return: the backquoted string
    :rtype: str
    """
    return javabridge.static_call(
        "Lweka/core/Utils;", "backQuoteChars", "(Ljava/lang/String;)Ljava/lang/String;", s)


def unbackquote(s):
    """
    Un-backquotes the string.

    :param s: the string to process
    :type s: str
    :return: the un-backquoted string
    :rtype: str
    """
    return javabridge.static_call(
        "Lweka/core/Utils;", "unbackQuoteChars", "(Ljava/lang/String;)Ljava/lang/String;", s)


def quote(s):
    """
    Quotes the string if necessary.

    :param s: the string to process
    :type s: str
    :return: the quoted string
    :rtype: str
    """
    return javabridge.static_call(
        "Lweka/core/Utils;", "quote", "(Ljava/lang/String;)Ljava/lang/String;", s)


def unquote(s):
    """
    Un-quotes the string.

    :param s: the string to process
    :type s: str
    :return: the un-quoted string
    :rtype: str
    """
    return javabridge.static_call(
        "Lweka/core/Utils;", "unquote", "(Ljava/lang/String;)Ljava/lang/String;", s)


def to_commandline(optionhandler):
    """
    Generates a commandline string from the OptionHandler instance.

    :param optionhandler: the OptionHandler instance to turn into a commandline
    :type optionhandler: OptionHandler
    :return: the commandline string
    :rtype: str
    """
    return javabridge.static_call(
        "Lweka/core/Utils;", "toCommandLine",
        "(Ljava/lang/Object;)Ljava/lang/String;",
        optionhandler.jobject)


def from_commandline(cmdline, classname=None):
    """
    Creates an OptionHandler based on the provided commandline string.

    :param cmdline: the commandline string to use
    :type cmdline: str
    :param classname: the classname of the wrapper to return other than OptionHandler (in dot-notation)
    :type classname: str
    :return: the generated option handler instance
    :rtype: object
    """
    params = split_options(cmdline)
    cls = params[0]
    params = params[1:]
    handler = OptionHandler(javabridge.static_call(
        "Lweka/core/Utils;", "forName",
        "(Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/String;)Ljava/lang/Object;",
        javabridge.class_for_name("java.lang.Object"), cls, params))
    if classname is None:
        return handler
    else:
        c = get_class(classname)
        return c(jobject=handler.jobject)


def complete_classname(classname):
    """
    Attempts to complete a partial classname like '.J48' and returns the full
    classname if a single match was found, otherwise an exception is raised.

    :param classname: the partial classname to expand
    :type classname: str
    :return: the full classname
    :rtype: str
    """

    result = javabridge.get_collection_wrapper(
        javabridge.static_call(
            "Lweka/Run;", "findSchemeMatch",
            "(Ljava/lang/String;Z)Ljava/util/List;",
            classname, True))
    if len(result) == 1:
        return str(result[0])
    elif len(result) == 0:
        raise Exception("No classname matches found for: " + classname)
    else:
        matches = []
        for i in range(len(result)):
            matches.append(str(result[i]))
        raise Exception("Found multiple matches for '" + classname + "':\n" + '\n'.join(matches))


class AbstractParameter(OptionHandler):
    """
    Ancestor for all parameter classes used by SetupGenerator and MultiSearch.
    """

    def __init__(self, classname=None, jobject=None, options=None):
        """
        Initializes the specified parameter using either the classname or the supplied JB_Object.

        :param classname: the classname of the search parameter
        :type classname: str
        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        if jobject is None:
            jobject = AbstractParameter.new_instance(classname)
        self.enforce_type(jobject, "weka.core.setupgenerator.AbstractParameter")
        super(AbstractParameter, self).__init__(jobject=jobject, options=options)

    @property
    def prop(self):
        """
        Returns the currently set property to apply the parameter to.

        :return: the property
        :rtype: str
        """
        return javabridge.call(self.jobject, "getProperty", "()Ljava/lang/String;")

    @prop.setter
    def prop(self, s):
        """
        Sets the property to apply the parameter to.

        :param s: the property
        :type s: str
        """
        javabridge.call(self.jobject, "setProperty", "(Ljava/lang/String;)V", s)


class ListParameter(AbstractParameter):
    """
    Parameter using a predefined list of values, used by SetupGenerator and MultiSearch.
    """

    def __init__(self, jobject=None, options=None):
        """
        Initializes the specified parameter using either its classname or the supplied JB_Object.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        classname = "weka.core.setupgenerator.ListParameter"
        if jobject is None:
            jobject = ListParameter.new_instance(classname)
        self.enforce_type(jobject, "weka.core.setupgenerator.ListParameter")
        super(ListParameter, self).__init__(jobject=jobject, options=options)

    @property
    def values(self):
        """
        Returns the currently set values.

        :return: the list of values (strings)
        :rtype: list
        """
        return split_options(javabridge.call(self.jobject, "getList", "()Ljava/lang/String;"))

    @values.setter
    def values(self, l):
        """
        Sets the list of values to apply.

        :param l: the list of values (strings)
        :type l: list
        """
        javabridge.call(self.jobject, "setList", "(Ljava/lang/String;)V", join_options(l))


class MathParameter(AbstractParameter):
    """
    Parameter using a math expression for generating values, used by SetupGenerator and MultiSearch.
    """

    def __init__(self, jobject=None, options=None):
        """
        Initializes the specified parameter using either its classname or the supplied JB_Object.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        classname = "weka.core.setupgenerator.MathParameter"
        if jobject is None:
            jobject = ListParameter.new_instance(classname)
        self.enforce_type(jobject, "weka.core.setupgenerator.MathParameter")
        super(MathParameter, self).__init__(jobject=jobject, options=options)

    @property
    def minimum(self):
        """
        Returns the currently set minimum value.

        :return: the minimum
        :rtype: float
        """
        return javabridge.call(self.jobject, "getMin", "()D")

    @minimum.setter
    def minimum(self, m):
        """
        Sets the new minimum value.

        :param m: the minimum
        :type m: float
        """
        javabridge.call(self.jobject, "setMin", "(D)V", m)

    @property
    def maximum(self):
        """
        Returns the currently set maximum value.

        :return: the maximum
        :rtype: float
        """
        return javabridge.call(self.jobject, "getMax", "()D")

    @maximum.setter
    def maximum(self, m):
        """
        Sets the new maximum value.

        :param m: the maximum
        :type m: float
        """
        javabridge.call(self.jobject, "setMax", "(D)V", m)

    @property
    def step(self):
        """
        Returns the currently set step value.

        :return: the step
        :rtype: float
        """
        return javabridge.call(self.jobject, "getStep", "()D")

    @step.setter
    def step(self, s):
        """
        Sets the new step value.

        :param s: the step
        :type s: float
        """
        javabridge.call(self.jobject, "setStep", "(D)V", s)

    @property
    def base(self):
        """
        Returns the currently set base value.

        :return: the base
        :rtype: float
        """
        return javabridge.call(self.jobject, "getBase", "()D")

    @base.setter
    def base(self, b):
        """
        Sets the new base value.

        :param b: the base
        :type b: float
        """
        javabridge.call(self.jobject, "setBase", "(D)V", b)

    @property
    def expression(self):
        """
        Returns the currently set expression.

        :return: the expression
        :rtype: str
        """
        return javabridge.call(self.jobject, "getExpression", "()Ljava/lang/String;")

    @expression.setter
    def expression(self, e):
        """
        Sets the new expression.

        :param e: the expression
        :type e: str
        """
        javabridge.call(self.jobject, "setExpression", "(Ljava/lang/String;)V", e)


class SetupGenerator(OptionHandler):
    """
    Allows generation of large number of setups using parameter setups.
    """

    def __init__(self, jobject=None, options=None):
        """
        Initializes the specified setup generator.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        classname = "weka.core.SetupGenerator"
        if jobject is None:
            jobject = OptionHandler.new_instance(classname)
        else:
            self.enforce_type(jobject, classname)
        super(SetupGenerator, self).__init__(jobject=jobject, options=options)

    @property
    def base_object(self):
        """
        Returns the base object to apply the setups to.

        :return: the base object
        :rtype: JavaObject or OptionHandler
        """
        jobj = javabridge.call(self.jobject, "getBaseObject", "()Ljava/io/Serializable;")
        if OptionHandler.check_type(jobj, "weka.core.OptionHandler"):
            return OptionHandler(jobj)
        else:
            return JavaObject(jobj)

    @base_object.setter
    def base_object(self, obj):
        """
        Sets the base object to apply the setups to.

        :param obj: the object to use (must be serializable!)
        :type obj: JavaObject
        """
        if not obj.is_serializable:
            raise Exception("Base object must be serializable: " + obj.classname)
        javabridge.call(self.jobject, "setBaseObject", "(Ljava/io/Serializable;)V", obj.jobject)

    @property
    def parameters(self):
        """
        Returns the list of currently set search parameters.

        :return: the list of AbstractSearchParameter objects
        :rtype: list
        """
        array = JavaArray(javabridge.call(self.jobject, "getParameters", "()[Lweka/core/setupgenerator/AbstractParameter;"))
        result = []
        for item in array:
            result.append(AbstractParameter(jobject=item.jobject))
        return result

    @parameters.setter
    def parameters(self, params):
        """
        Sets the list of search parameters to use.

        :param params: list of AbstractSearchParameter objects
        :type params: list
        """
        array = JavaArray(jobject=JavaArray.new_instance("weka.core.setupgenerator.AbstractParameter", len(params)))
        for idx, obj in enumerate(params):
            array[idx] = obj.jobject
        javabridge.call(self.jobject, "setParameters", "([Lweka/core/setupgenerator/AbstractParameter;)V", array.jobject)

    def setups(self):
        """
        Generates and returns all the setups according to the parameter search space.

        :return: the list of configured objects (of type JavaObject)
        :rtype: list
        """
        result = []
        has_options = self.base_object.is_optionhandler
        enm = javabridge.get_enumeration_wrapper(javabridge.call(self.jobject, "setups", "()Ljava/util/Enumeration;"))
        while enm.hasMoreElements():
            if has_options:
                result.append(OptionHandler(enm.nextElement()))
            else:
                result.append(JavaObject(enm.nextElement()))
        return result


class Date(JavaObject):
    """
    Wraps a java.util.Date object.
    """

    def __init__(self, jobject=None, msecs=None):
        """
        Initializes the specified Date. Uses the current timestamp if no jobject or msecs provided.

        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param msecs: the milli-seconds since 1970
        :type msecs: long
        """
        if jobject is None:
            if msecs is None:
                jobject = javabridge.make_instance("java/util/Date", "()V")
            else:
                jobject = javabridge.make_instance("java/util/Date", "(J)V", msecs)
        else:
            self.enforce_type(jobject, "java.util.Date")
        super(Date, self).__init__(jobject=jobject)

    @property
    def time(self):
        """
        Returns the stored milli-seconds.

        :return: the milli-seconds
        :rtype: long
        """
        return javabridge.call(self.jobject, "getTime", "()J")

    @time.setter
    def time(self, msecs):
        """
        Sets the milli-seconds.

        :param msecs: the milli-seconds to use
        :return:
        """
        javabridge.call(self.jobject, "setTime", "(J)V", msecs)


def load_suggestions():
    """
    Loads the class/package suggestions, if necessary.
    """
    global suggestions
    if suggestions is not None:
        return
    filename = jvm.lib_dir() + os.sep + "pkg_suggestions.csv"
    suggestions = {}
    with open(filename) as csvfile:
        csvreader = csv.reader(csvfile)
        for row in csvreader:
            if len(row[1]) > 0:
                suggestions[row[1]] = row[0]


def suggest_package(name, exact=False):
    """
    Suggests package(s) for the given name (classname, package name). Matching can be either exact or just a substring.

    :param name: the name to look for
    :type name: str
    :param exact: whether to perform exact matching or substring matching
    :type exact: bool
    :return: list of matching package names
    :rtype: list
    """

    global suggestions

    result = []

    load_suggestions()

    for cname in suggestions:
        add = False
        if exact:
            if name == cname:
                add = True
        else:
            if name in cname:
                add = True
        if add:
            pkg = suggestions[cname]
            if pkg not in result:
                result.append(pkg)

    return result


def main():
    """
    Runs a classifier from the command-line. Calls JVM start/stop automatically.
    Use -h to see all options.
    """
    parser = argparse.ArgumentParser(
        description='Performs option handling operations from the command-line. Calls JVM start/stop automatically.')
    parser.add_argument("-j", metavar="classpath", dest="classpath", help="additional classpath, jars/directories")
    parser.add_argument("-action", metavar="action", dest="action", required=True,
                        help="The action to perform on the options: join=create single string, "\
                             "split=create array from quoted option, code=generate code from options")
    parser.add_argument("option", nargs=argparse.REMAINDER, help="The option(s) to process")
    parsed = parser.parse_args()
    jars = []
    if parsed.classpath is not None:
        jars = parsed.classpath.split(os.pathsep)

    jvm.start(jars, packages=True)

    try:
        if parsed.action == "join":
            output = "cmdline = \"" + backquote(join_options(parsed.option)) + "\""
        elif parsed.action == "split" and len(parsed.option) == 1:
            output = "options = [\n"
            opts = split_options(parsed.option[0])
            for idx, opt in enumerate(opts):
                if idx > 0:
                    output += ",\n"
                output += "    \"" + backquote(opt) + "\""
            output += "]"
        elif parsed.action == "code":
            options = parsed.option[:]
            cname = None
            # classname + options?
            if options[0].find(".") > -1:
                cname = options[0]
                options = options[1:]
            output = "options = [ \n"
            for idx, opt in enumerate(options):
                if idx > 0:
                    output += ",\n"
                output += "    \"" + backquote(opt) + "\""
            output += "]\n"
            if cname is not None:
                output += 'handler = OptionHandler(JavaObject.new_instance("' + cname + '"))\n'
                output += 'handler.options = options\n'
        else:
            raise Exception("Unsupported action: " + parsed.action)
        if output is not None:
            print(output)
    except Exception as e:
        print(e)
    finally:
        jvm.stop()

if __name__ == "__main__":
    try:
        main()
    except Exception as ex:
        print(ex)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# associations.py
# Copyright (C) 2014-2021 Fracpete (pythonwekawrapper at gmail dot com)

import javabridge
import logging
import os
import sys
import argparse
import traceback
import weka.core.jvm as jvm
import weka.core.converters as converters
from weka.core.classes import OptionHandler, JavaObject, join_options
from weka.core.capabilities import Capabilities
from weka.core.dataset import Attribute
from weka.core.typeconv import jstring_array_to_list

# logging setup
logger = logging.getLogger("weka.associations")


class Item(JavaObject):
    """
    Wrapper for weka.associations.Item class.
    """

    def __init__(self, jobject):
        """
        Initializes the wrapper.

        :param jobject: the Item object to wrap
        :type jobject: JB_Object
        """

        if jobject is None:
            raise Exception("No Item JB_Object provided!")
        self.enforce_type(jobject, "weka.associations.Item")
        super(Item, self).__init__(jobject)

    def __compareto__(self, other):
        """
        Return the comparison result with the other rule.

        :param other: the other rule to compare with
        :type other: Item
        :return: the integer result, <0 is less than, =0 is equal to, >0 is greater than
        :rtype: int
        """
        return javabridge.call(self.jobject, "compareTo", "(Lweka/associations/Item;)I", other.jobject)

    def __lt__(self, other):
        """
        Checks self < other.

        :param other: the item to compare with
        :type other: Item
        :return: whether less than the other item
        :rtype: bool
        """
        return self.__compareto__(other) < 0

    def __le__(self, other):
        """
        Checks self <= other.

        :param other: the item to compare with
        :type other: Item
        :return: whether less or equal than the other item
        :rtype: bool
        """
        return self.__compareto__(other) <= 0

    def decrease_frequency(self, frequency=None):
        """
        Decreases the frequency.

        :param frequency: the frequency to decrease by, 1 if None
        :type frequency: int
        """
        if frequency is None:
            javabridge.call(self.jobject, "decreaseFrequency", "()V")
        else:
            javabridge.call(self.jobject, "decreaseFrequency", "(I)V", frequency)

    def increase_frequency(self, frequency=None):
        """
        Increases the frequency.

        :param frequency: the frequency to increase by, 1 if None
        :type frequency: int
        """
        if frequency is None:
            javabridge.call(self.jobject, "increaseFrequency", "()V")
        else:
            javabridge.call(self.jobject, "increaseFrequency", "(I)V", frequency)

    def __eq__(self, other):
        """
        Returns whether this object is the same as the other one.

        :param other: the object to compare with
        :type other: Item
        :return: True if the same
        :rtype: bool
        """
        return javabridge.call(self.jobject, "equals", "(Ljava/lang/Object;)Z", other.jobject)

    def __ge__(self, other):
        """
        Checks self >= other.

        :param other: the item to compare with
        :type other: Item
        :return: whether greater or equal than the other item
        :rtype: bool
        """
        return self.__compareto__(other) >= 0

    def __gt__(self, other):
        """
        Checks self > other.

        :param other: the item to compare with
        :type other: Item
        :return: whether greater than the other item
        :rtype: bool
        """
        return self.__compareto__(other) > 0

    def __str__(self):
        """
        Returns the item as string.

        :return: the item as string
        :rtype: str
        """
        return javabridge.call(self.jobject, "toString", "()Ljava/lang/String;")

    @property
    def frequency(self):
        """
        Returns the frequency.

        :return: the frequency
        :rtype: int
        """
        return javabridge.call(self.jobject, "getFrequency", "()I")

    @property
    def attribute(self):
        """
        Returns the attribute.

        :return: the attribute
        :rtype: Attribute
        """
        return Attribute(javabridge.call(self.jobject, "getAttribute", "()Lweka/core/Attribute;"))

    @property
    def comparison(self):
        """
        Returns the comparison operator as string.

        :return: the comparison iterator
        :rtype: str
        """
        return Attribute(javabridge.call(self.jobject, "getComparisonAsString", "()Ljava/lang/String;"))

    @property
    def item_value(self):
        """
        Returns the item value as string.

        :return: the item value
        :rtype: str
        """
        return Attribute(javabridge.call(self.jobject, "getItemValueAsString", "()Ljava/lang/String;"))


class AssociationRule(JavaObject):
    """
    Wrapper for weka.associations.AssociationRule class.
    """

    def __init__(self, jobject):
        """
        Initializes the wrapper.

        :param jobject: the AssociationRule object to wrap
        :type jobject: JB_Object
        """
        if jobject is None:
            raise Exception("No AssociationRule JB_Object provided!")
        self.enforce_type(jobject, "weka.associations.AssociationRule")
        super(AssociationRule, self).__init__(jobject)

    def __compareto__(self, other):
        """
        Return the comparison result with the other rule.

        :param other: the other rule to compare with
        :type other: AssociationRule
        :return: the integer result, <0 is less than, =0 is equal to, >0 is greater than
        :rtype: int
        """
        return javabridge.call(self.jobject, "compareTo", "(Lweka/associations/AssociationRule;)I", other.jobject)

    def __lt__(self, other):
        """
        Checks self < other.

        :param other: the rule to compare with
        :type other: AssociationRule
        :return: whether less than the other rule
        :rtype: bool
        """
        return self.__compareto__(other) < 0

    def __le__(self, other):
        """
        Checks self <= other.

        :param other: the rule to compare with
        :type other: AssociationRule
        :return: whether less or equal than the other rule
        :rtype: bool
        """
        return self.__compareto__(other) <= 0

    def __eq__(self, other):
        """
        Returns whether this object is the same as the other one.

        :param other: the object to compare with
        :type other: AssociationRule
        :return: True if the same
        :rtype: bool
        """
        return javabridge.call(self.jobject, "equals", "(Ljava/lang/Object;)Z", other.jobject)

    def __ge__(self, other):
        """
        Checks self >= other.

        :param other: the rule to compare with
        :type other: AssociationRule
        :return: whether greater or equal than the other rule
        :rtype: bool
        """
        return self.__compareto__(other) >= 0

    def __gt__(self, other):
        """
        Checks self > other.

        :param other: the rule to compare with
        :type other: AssociationRule
        :return: whether greater than the other rule
        :rtype: bool
        """
        return self.__compareto__(other) > 0

    @property
    def consequence(self):
        """
        Get the the consequence.

        :return: the consequence, list of Item objects
        :rtype: list
        """
        items = javabridge.get_collection_wrapper(
            javabridge.call(self.jobject, "getConsequence", "()Ljava/util/Collection;"))
        result = []
        for item in items:
            result.append(Item(item))
        return result

    @property
    def consequence_support(self):
        """
        Get the support for the consequence.

        :return: the support
        :rtype: int
        """
        return javabridge.call(self.jobject, "getConsequenceSupport", "()I")

    @property
    def premise(self):
        """
        Get the the premise.

        :return: the premise, list of Item objects
        :rtype: list
        """
        items = javabridge.get_collection_wrapper(
            javabridge.call(self.jobject, "getPremise", "()Ljava/util/Collection;"))
        result = []
        for item in items:
            result.append(Item(item))
        return result

    @property
    def premise_support(self):
        """
        Get the support for the premise.

        :return: the support
        :rtype: int
        """
        return javabridge.call(self.jobject, "getPremiseSupport", "()I")

    @property
    def total_support(self):
        """
        Get the total support.

        :return: the support
        :rtype: int
        """
        return javabridge.call(self.jobject, "getTotalSupport", "()I")

    @property
    def total_transactions(self):
        """
        Get the total transactions.

        :return: the transactions
        :rtype: int
        """
        return javabridge.call(self.jobject, "getTotalTransactions", "()I")

    @property
    def metric_names(self):
        """
        Returns the metric names for the rule.

        :return: the metric names
        :rtype: list
        """
        return jstring_array_to_list(javabridge.call(self.jobject, "getMetricNamesForRule", "()[Ljava/lang/String;"))

    @property
    def metric_values(self):
        """
        Returns the metric values for the rule.

        :return: the metric values
        :rtype: ndarray
        """
        return javabridge.get_env().get_double_array_elements(
            javabridge.call(self.jobject, "getMetricValuesForRule", "()[D"))

    def metric_value(self, name):
        """
        Returns the named metric value for the rule.

        :param name: the name of the metric
        :type name: str
        :return: the metric value
        :rtype: float
        """
        return javabridge.call(self.jobject, "getNamedMetricValue", "(Ljava/lang/String;)D", name)

    @property
    def primary_metric_name(self):
        """
        Returns the primary metric name for the rule.

        :return: the metric name
        :rtype: str
        """
        return javabridge.call(self.jobject, "getPrimaryMetricName", "()Ljava/lang/String;")

    @property
    def primary_metric_value(self):
        """
        Returns the primary metric value for the rule.

        :return: the metric value
        :rtype: float
        """
        return javabridge.call(self.jobject, "getPrimaryMetricValue", "()D")
    
    def to_dict(self):
        """
        Builds a dictionary with the properties of the AssociationRule object.

        :return: the AssociationRule dictionary
        :rtype: dict
        """
        rule = {
            'premise': self.premise,
            'premise_support': self.premise_support,
            'consequence': self.consequence,
            'consequence_support': self.consequence_support,
            'total_support': self.total_support
        }
        index_names = [x.lower() for x in self.metric_names]
        metrics = dict(zip(index_names, self.metric_values))
        rule.update(metrics)
        return rule


class AssociationRulesIterator(object):
    """
    Iterator for weka.associations.AssociationRules class.
    """

    def __init__(self, rules):
        """
        Initializes with the rules.

        :param rules: the rules to use
        :type rules: AssociationRules
        """
        self.rules = rules
        self.index = 0
        self.length = len(rules)

    def __iter__(self):
        """
        Returns itself.
        """
        return self

    def __next__(self):
        """
        Returns the next rule.

        :return: the next rule object
        :rtype: AssociationRule
        """
        if self.index < self.length:
            index = self.index
            self.index += 1
            return self.rules[index]
        else:
            raise StopIteration()


class AssociationRules(JavaObject):
    """
    Wrapper for weka.associations.AssociationRules class.
    """

    def __init__(self, jobject):
        """
        Initializes the wrapper.

        :param jobject: the AssociationRules object to wrap
        :type jobject: JB_Object
        """
        if jobject is None:
            raise Exception("No AssociationRules JB_Object provided!")
        self.enforce_type(jobject, "weka.associations.AssociationRules")
        super(AssociationRules, self).__init__(jobject)

    def __len__(self):
        """
        Returns the number of rules available.

        :return: the number of rules
        :rtype: int
        """
        return javabridge.call(self.jobject, "getNumRules", "()I")

    def __getitem__(self, item):
        """
        Returns the specified item, using 0-based indices.

        :param item: the 0-based index
        :type item: int
        :return: the AssociationRule
        :rtype: AssociationRule
        """
        rules = javabridge.get_collection_wrapper(
            javabridge.call(self.jobject, "getRules", "()Ljava/util/List;"))
        return AssociationRule(rules[item])

    def __str__(self):
        """
        Returns a short description of the rules.

        :return: short description
        :rtype: str
        """
        return str(len(self)) + " rules generated by " + self.producer

    def __iter__(self):
        """
        Returns an iterator for the rules.

        :return: the iterator
        :rtype: AssociationRulesIterator
        """
        return AssociationRulesIterator(self)

    @property
    def producer(self):
        """
        Returns a string describing the producer that generated these rules.

        :return: the producer
        :rtype: str
        """
        return javabridge.call(self.jobject, "getProducer", "()Ljava/lang/String;")

    @producer.setter
    def producer(self, producer):
        """"
        Sets the string describing the producer that generated these rules.

        :param producer: the producer
        :type producer: str
        """
        javabridge.call(self.jobject, "setProducer", "(Ljava/lang/String;)V", producer)
    
    def to_dict(self):
        """
        Returns a list of association rules in dict format

        :return: the association rules
        :rtype: list
        """
        rules = []
        for rule in self:
            rules.append(rule.to_dict())
        return rules


class Associator(OptionHandler):
    """
    Wrapper class for associators.
    """

    def __init__(self, classname=None, jobject=None, options=None):
        """
        Initializes the specified associator using either the classname or the supplied JB_Object.

        :param classname: the classname of the associator
        :type classname: str
        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        if jobject is None:
            jobject = Associator.new_instance(classname)
        self.enforce_type(jobject, "weka.associations.Associator")
        super(Associator, self).__init__(jobject=jobject, options=options)

    @property
    def capabilities(self):
        """
        Returns the capabilities of the associator.

        :return: the capabilities
        :rtype: Capabilities
        """
        return Capabilities(javabridge.call(self.jobject, "getCapabilities", "()Lweka/core/Capabilities;"))

    def build_associations(self, data):
        """
        Builds the associator with the data.

        :param data: the data to train the associator with
        :type data: Instances
        """
        javabridge.call(self.jobject, "buildAssociations", "(Lweka/core/Instances;)V", data.jobject)

    def can_produce_rules(self):
        """
        Checks whether association rules can be generated.

        :return: whether scheme implements AssociationRulesProducer interface and association rules can be generated
        :rtype: bool
        """
        if not self.check_type(self.jobject, "weka.associations.AssociationRulesProducer"):
            return False
        return javabridge.call(self.jobject, "canProduceRules", "()Z")

    def association_rules(self):
        """
        Returns association rules that were generated. Only if implements AssociationRulesProducer.

        :return: the association rules that were generated
        :rtype: AssociationRules
        """
        if not self.check_type(self.jobject, "weka.associations.AssociationRulesProducer"):
            return None
        return AssociationRules(
            javabridge.call(self.jobject, "getAssociationRules", "()Lweka/associations/AssociationRules;"))

    @property
    def rule_metric_names(self):
        """
        Returns the rule metric names of the association rules. Only if implements AssociationRulesProducer.

        :return: the metric names
        :rtype: list
        """
        if not self.check_type(self.jobject, "weka.associations.AssociationRulesProducer"):
            return None
        return jstring_array_to_list(
            javabridge.call(self.jobject, "getRuleMetricNames", "()[Ljava/lang/String;"))

    @classmethod
    def make_copy(cls, associator):
        """
        Creates a copy of the clusterer.

        :param associator: the associator to copy
        :type associator: Associator
        :return: the copy of the associator
        :rtype: Associator
        """
        return Associator(
            jobject=javabridge.static_call(
                "weka/associations/AbstractAssociator", "makeCopy",
                "(Lweka/associations/Associator;)Lweka/associations/Associator;", associator.jobject))


def main(args=None):
    """
    Runs a associator from the command-line. Calls JVM start/stop automatically.
    Use -h to see all options.

    :param args: the command-line arguments to use, uses sys.argv if None
    :type args: list
    """

    parser = argparse.ArgumentParser(
        description='Executes an associator from the command-line. Calls JVM start/stop automatically.')
    parser.add_argument("-j", metavar="classpath", dest="classpath", help="additional classpath, jars/directories")
    parser.add_argument("-X", metavar="heap", dest="heap", help="max heap size for jvm, e.g., 512m")
    parser.add_argument("-t", metavar="train", dest="train", required=True, help="training set file")
    parser.add_argument("associator", help="associator classname, e.g., weka.associations.Apriori")
    parser.add_argument("option", nargs=argparse.REMAINDER, help="additional associator options")
    parsed = parser.parse_args(args=args)
    jars = []
    if parsed.classpath is not None:
        jars = parsed.classpath.split(os.pathsep)

    jvm.start(jars, max_heap_size=parsed.heap, packages=True)

    logger.debug("Commandline: " + join_options(sys.argv[1:]))

    try:
        associator = Associator(classname=parsed.associator)
        if len(parsed.option) > 0:
            associator.options = parsed.option
        loader = converters.loader_for_file(parsed.train)
        data = loader.load_file(parsed.train)
        associator.build_associations(data)
        print(str(associator))
    except Exception:
        print(traceback.format_exc())
    finally:
        jvm.stop()


def sys_main():
    """
    Runs the main function using the system cli arguments, and
    returns a system error code.

    :return: 0 for success, 1 for failure.
    :rtype: int
    """

    try:
        main()
        return 0
    except Exception:
        print(traceback.format_exc())
        return 1


if __name__ == "__main__":
    try:
        main()
    except Exception:
        print(traceback.format_exc())



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# experiments.py
# Copyright (C) 2014-2016 Fracpete (pythonwekawrapper at gmail dot com)

import logging
import javabridge
from weka.core.classes import OptionHandler, Range, from_commandline
from weka.core.dataset import Instances
from weka.classifiers import Classifier

# logging setup
logger = logging.getLogger("weka.experiments")


class Experiment(OptionHandler):
    """
    Wrapper class for an experiment.
    """

    def __init__(self, classname="weka.experiment.Experiment", jobject=None, options=None):
        """
        Initializes the specified experiment using either the classname or the supplied JB_Object.

        :param classname: the classname of the experiment
        :type classname: str
        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to use
        :type options: list
        """
        if jobject is None:
            jobject = Experiment.new_instance(classname)
        super(Experiment, self).__init__(jobject=jobject, options=options)


class SimpleExperiment(OptionHandler):
    """
    Ancestor for simple experiments.

    See following URL for how to use the Experiment API:
    http://weka.wikispaces.com/Using+the+Experiment+API
    """

    def __init__(self, datasets, classifiers, jobject=None, classification=True, runs=10, result=None):
        """
        Initializes the experiment.

        :param datasets: the filenames of datasets to use in the experiment
        :type datasets: list
        :param classifiers: the Classifier objects or commandline strings to use in the experiment
        :type classifiers: list
        :param jobject: the Java Object to use
        :type jobject: JB_Object
        :param classification: whether to perform classification or regression
        :type classification:bool
        :param runs: the number of runs to perform
        :type runs: int
        :param result: the filename of the file to store the results in
        :type result: str
        """

        if not jobject is None:
            self.enforce_type(jobject, "weka.experiment.Experiment")
        if jobject is None:
            jobject = javabridge.make_instance("weka/experiment/Experiment", "()V")

        self.classification = classification
        self.runs = runs
        self.datasets = datasets[:]
        self.classifiers = classifiers[:]
        self.result = result
        super(SimpleExperiment, self).__init__(jobject=jobject)

    def configure_splitevaluator(self):
        """
        Configures and returns the SplitEvaluator and Classifier instance as tuple.

        :return: evaluator and classifier
        :rtype: tuple
        """
        if self.classification:
            speval = javabridge.make_instance("weka/experiment/ClassifierSplitEvaluator", "()V")
        else:
            speval = javabridge.make_instance("weka/experiment/RegressionSplitEvaluator", "()V")
        classifier = javabridge.call(speval, "getClassifier", "()Lweka/classifiers/Classifier;")
        return speval, classifier

    def configure_resultproducer(self):
        """
        Configures and returns the ResultProducer and PropertyPath as tuple.

        :return: producer and property path
        :rtype: tuple
        """
        raise Exception("Not implemented!")

    def setup(self):
        """
        Initializes the experiment.
        """
        # basic options
        javabridge.call(
            self.jobject, "setPropertyArray", "(Ljava/lang/Object;)V",
            javabridge.get_env().make_object_array(0, javabridge.get_env().find_class("weka/classifiers/Classifier")))
        javabridge.call(
            self.jobject, "setUsePropertyIterator", "(Z)V", True)
        javabridge.call(
            self.jobject, "setRunLower", "(I)V", 1)
        javabridge.call(
            self.jobject, "setRunUpper", "(I)V", self.runs)

        # setup result producer
        rproducer, prop_path = self.configure_resultproducer()
        javabridge.call(
            self.jobject, "setResultProducer", "(Lweka/experiment/ResultProducer;)V", rproducer)
        javabridge.call(
            self.jobject, "setPropertyPath", "([Lweka/experiment/PropertyNode;)V", prop_path)

        # classifiers
        classifiers = javabridge.get_env().make_object_array(
            len(self.classifiers), javabridge.get_env().find_class("weka/classifiers/Classifier"))
        for i, classifier in enumerate(self.classifiers):
            if type(classifier) is Classifier:
                javabridge.get_env().set_object_array_element(
                    classifiers, i, classifier.jobject)
            else:
                javabridge.get_env().set_object_array_element(
                    classifiers, i, from_commandline(classifier).jobject)
        javabridge.call(
            self.jobject, "setPropertyArray", "(Ljava/lang/Object;)V",
            classifiers)

        # datasets
        datasets = javabridge.make_instance("javax/swing/DefaultListModel", "()V")
        for dataset in self.datasets:
            f = javabridge.make_instance("java/io/File", "(Ljava/lang/String;)V", dataset)
            javabridge.call(datasets, "addElement", "(Ljava/lang/Object;)V", f)
        javabridge.call(
            self.jobject, "setDatasets", "(Ljavax/swing/DefaultListModel;)V", datasets)

        # output file
        if str(self.result).lower().endswith(".arff"):
            rlistener = javabridge.make_instance("weka/experiment/InstancesResultListener", "()V")
        elif str(self.result).lower().endswith(".csv"):
            rlistener = javabridge.make_instance("weka/experiment/CSVResultListener", "()V")
        else:
            raise Exception("Unhandled output format for results: " + self.result)
        rfile = javabridge.make_instance("java/io/File", "(Ljava/lang/String;)V", self.result)
        javabridge.call(
            rlistener, "setOutputFile", "(Ljava/io/File;)V", rfile)
        javabridge.call(
            self.jobject, "setResultListener", "(Lweka/experiment/ResultListener;)V", rlistener)

    def run(self):
        """
        Executes the experiment.
        """
        logger.info("Initializing...")
        javabridge.call(self.jobject, "initialize", "()V")
        logger.info("Running...")
        javabridge.call(self.jobject, "runExperiment", "()V")
        logger.info("Finished...")
        javabridge.call(self.jobject, "postProcess", "()V")

    def experiment(self):
        """
        Returns the internal experiment, if set up, otherwise None.

        :return: the internal experiment
        :rtype: Experiment
        """
        if self.jobject is None:
            return None
        else:
            return Experiment(jobject=self.jobject)

    @classmethod
    def load(cls, filename):
        """
        Loads the experiment from disk.

        :param filename: the filename of the experiment to load
        :type filename: str
        :return: the experiment
        :rtype: Experiment
        """
        jobject = javabridge.static_call(
            "weka/experiment/Experiment", "read", "(Ljava/lang/String;)Lweka/experiment/Experiment;",
            filename)
        return Experiment(jobject=jobject)

    @classmethod
    def save(cls, filename, experiment):
        """
        Saves the experiment to disk.

        :param filename: the filename to save the experiment to
        :type filename: str
        :param experiment: the Experiment to save
        :type experiment: Experiment
        """
        javabridge.static_call(
            "weka/experiment/Experiment", "write", "(Ljava/lang/String;Lweka/experiment/Experiment;)V",
            filename, experiment.jobject)


class SimpleCrossValidationExperiment(SimpleExperiment):
    """
    Performs a simple cross-validation experiment. Can output the results either in ARFF or CSV.
    """

    def __init__(self, datasets, classifiers, classification=True, runs=10, folds=10, result=None):
        """
        Initializes the experiment.

        :param datasets: the filenames of datasets to use in the experiment
        :type datasets: list
        :param classifiers: the Classifier objects to use in the experiment
        :type classifiers: list
        :param classification: whether to perform classification
        :type classification: bool
        :param runs: the number of runs to perform
        :type runs: int
        :param folds: the number folds to use for CV
        :type folds: int
        :param result: the filename of the file to store the results in
        :type result: str
        """

        if runs < 1:
            raise Exception("Number of runs must be at least 1!")
        if folds < 2:
            raise Exception("Number of folds must be at least 2!")
        if len(datasets) == 0:
            raise Exception("No datasets provided!")
        if len(classifiers) == 0:
            raise Exception("No classifiers provided!")
        if result is None:
            raise Exception("No filename for results provided!")

        super(SimpleCrossValidationExperiment, self).__init__(
            classification=classification, runs=runs, datasets=datasets,
            classifiers=classifiers, result=result)

        self.folds = folds

    def configure_resultproducer(self):
        """
        Configures and returns the ResultProducer and PropertyPath as tuple.

        :return: producer and property path
        :rtype: tuple
        """
        rproducer = javabridge.make_instance("weka/experiment/CrossValidationResultProducer", "()V")
        javabridge.call(rproducer, "setNumFolds", "(I)V", self.folds)
        speval, classifier = self.configure_splitevaluator()
        javabridge.call(rproducer, "setSplitEvaluator", "(Lweka/experiment/SplitEvaluator;)V", speval)
        prop_path = javabridge.get_env().make_object_array(
            2, javabridge.get_env().find_class("weka/experiment/PropertyNode"))
        cls = javabridge.get_env().find_class("weka/experiment/CrossValidationResultProducer")
        desc = javabridge.make_instance(
            "java/beans/PropertyDescriptor", "(Ljava/lang/String;Ljava/lang/Class;)V", "splitEvaluator", cls)
        node = javabridge.make_instance(
            "weka/experiment/PropertyNode", "(Ljava/lang/Object;Ljava/beans/PropertyDescriptor;Ljava/lang/Class;)V",
            speval, desc, cls)
        javabridge.get_env().set_object_array_element(prop_path, 0, node)
        cls = javabridge.get_env().get_object_class(speval)
        desc = javabridge.make_instance(
            "java/beans/PropertyDescriptor", "(Ljava/lang/String;Ljava/lang/Class;)V", "classifier", cls)
        node = javabridge.make_instance(
            "weka/experiment/PropertyNode", "(Ljava/lang/Object;Ljava/beans/PropertyDescriptor;Ljava/lang/Class;)V",
            javabridge.call(speval, "getClass", "()Ljava/lang/Class;"), desc, cls)
        javabridge.get_env().set_object_array_element(prop_path, 1, node)

        return rproducer, prop_path


class SimpleRandomSplitExperiment(SimpleExperiment):
    """
    Performs a simple random split experiment. Can output the results either in ARFF or CSV.
    """

    def __init__(self, datasets, classifiers, classification=True, runs=10, percentage=66.6, preserve_order=False,
                 result=None):
        """
        Initializes the experiment.

        :param classification: whether to perform classification or regression
        :type classification: bool
        :param runs: the number of runs to perform
        :type runs: int
        :param percentage: the percentage to use for training
        :type percentage: float
        :param preserve_order: whether to preserve the order in the datasets
        :type preserve_order: bool
        :param datasets: the filenames of datasets to use in the experiment
        :type datasets: list
        :param classifiers: the Classifier objects to use in the experiment
        :type classifiers: list
        :param result: the filename of the file to store the results in
        :type result: str
        """

        if runs < 1:
            raise Exception("Number of runs must be at least 1!")
        if percentage <= 0:
            raise Exception("Percentage for training must be >0!")
        if percentage >= 100:
            raise Exception("Percentage for training must be <100!")
        if len(datasets) == 0:
            raise Exception("No datasets provided!")
        if len(classifiers) == 0:
            raise Exception("No classifiers provided!")
        if result is None:
            raise Exception("No filename for results provided!")

        super(SimpleRandomSplitExperiment, self).__init__(
            classification=classification, runs=runs, datasets=datasets,
            classifiers=classifiers, result=result)

        self.percentage = percentage
        self.preserve_order = preserve_order

    def configure_resultproducer(self):
        """
        Configures and returns the ResultProducer and PropertyPath as tuple.

        :return: producer and property path
        :rtype: tuple
        """
        rproducer = javabridge.make_instance("weka/experiment/RandomSplitResultProducer", "()V")
        javabridge.call(rproducer, "setRandomizeData", "(Z)V", not self.preserve_order)
        javabridge.call(rproducer, "setTrainPercent", "(D)V", self.percentage)
        speval, classifier = self.configure_splitevaluator()
        javabridge.call(rproducer, "setSplitEvaluator", "(Lweka/experiment/SplitEvaluator;)V", speval)
        prop_path = javabridge.get_env().make_object_array(
            2, javabridge.get_env().find_class("weka/experiment/PropertyNode"))
        cls = javabridge.get_env().find_class("weka/experiment/RandomSplitResultProducer")
        desc = javabridge.make_instance(
            "java/beans/PropertyDescriptor", "(Ljava/lang/String;Ljava/lang/Class;)V", "splitEvaluator", cls)
        node = javabridge.make_instance(
            "weka/experiment/PropertyNode", "(Ljava/lang/Object;Ljava/beans/PropertyDescriptor;Ljava/lang/Class;)V",
            speval, desc, cls)
        javabridge.get_env().set_object_array_element(prop_path, 0, node)
        cls = javabridge.get_env().get_object_class(speval)
        desc = javabridge.make_instance(
            "java/beans/PropertyDescriptor", "(Ljava/lang/String;Ljava/lang/Class;)V", "classifier", cls)
        node = javabridge.make_instance(
            "weka/experiment/PropertyNode", "(Ljava/lang/Object;Ljava/beans/PropertyDescriptor;Ljava/lang/Class;)V",
            javabridge.call(speval, "getClass", "()Ljava/lang/Class;"), desc, cls)
        javabridge.get_env().set_object_array_element(prop_path, 1, node)

        return rproducer, prop_path


class ResultMatrix(OptionHandler):
    """
    For generating results from an Experiment run.
    """

    def __init__(self, classname="weka.experiment.ResultMatrixPlainText", jobject=None, options=None):
        """
        Initializes the specified ResultMatrix using either the classname or the supplied JB_Object.

        :param classname: the classname of the ResultMatrix
        :type classname: str
        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to use
        :type options: list
        """
        if jobject is None:
            jobject = ResultMatrix.new_instance(classname)
        self.enforce_type(jobject, "weka.experiment.ResultMatrix")
        super(ResultMatrix, self).__init__(jobject=jobject, options=options)

    @property
    def rows(self):
        """
        Returns the row count.

        :return: the count
        :rtype: int
        """
        return javabridge.call(self.jobject, "getRowCount", "()I")

    @property
    def columns(self):
        """
        Returns the column count.

        :return: the count
        :rtype: int
        """
        return javabridge.call(self.jobject, "getColCount", "()I")

    def is_col_hidden(self, index):
        """
        Returns whether the column is hidden.

        :param index: the 0-based column index
        :type index: int
        :return: true if hidden
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getColHidden", "(I)Z", index)

    def hide_col(self, index):
        """
        Hides the column.

        :param index: the 0-based column index
        :type index: int
        """
        javabridge.call(self.jobject, "setColHidden", "(IZ)V", index, True)

    def show_col(self, index):
        """
        Shows the column.

        :param index: the 0-based column index
        :type index: int
        """
        javabridge.call(self.jobject, "setColHidden", "(IZ)V", index, False)

    def is_row_hidden(self, index):
        """
        Returns whether the row is hidden.

        :param index: the 0-based row index
        :type index: int
        :return: true if hidden
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getRowHidden", "(I)Z", index)

    def hide_row(self, index):
        """
        Hides the row.

        :param index: the 0-based row index
        :type index: int
        """
        javabridge.call(self.jobject, "setRowHidden", "(IZ)V", index, True)

    def show_row(self, index):
        """
        Shows the row.

        :param index: the 0-based row index
        :type index: int
        """
        javabridge.call(self.jobject, "setRowHidden", "(IZ)V", index, False)

    def get_row_name(self, index):
        """
        Returns the row name.

        :param index: the 0-based row index
        :type index: int
        :return: the row name, None if invalid index
        :rtype: str
        """
        return javabridge.call(self.jobject, "getRowName", "(I)Ljava/lang/String;", index)

    def set_row_name(self, index, name):
        """
        Sets the row name.

        :param index: the 0-based row index
        :type index: int
        :param name: the name of the row
        :type name: str
        """
        javabridge.call(self.jobject, "setRowName", "(ILjava/lang/String;)V", index, name)

    def get_col_name(self, index):
        """
        Returns the column name.

        :param index: the 0-based row index
        :type index: int
        :return: the column name, None if invalid index
        :rtype: str
        """
        return javabridge.call(self.jobject, "getColName", "(I)Ljava/lang/String;", index)

    def set_col_name(self, index, name):
        """
        Sets the column name.

        :param index: the 0-based row index
        :type index: int
        :param name: the name of the column
        :type name: str
        """
        javabridge.call(self.jobject, "setColName", "(ILjava/lang/String;)V", index, name)

    def get_mean(self, col, row):
        """
        Returns the mean at this location (if valid location).

        :param col: the 0-based column index
        :type col: int
        :param row: the 0-based row index
        :type row: int
        :return: the mean
        :rtype: float
        """
        return javabridge.call(self.jobject, "getMean", "(II)D", col, row)

    def set_mean(self, col, row, mean):
        """
        Sets the mean at this location (if valid location).

        :param col: the 0-based column index
        :type col: int
        :param row: the 0-based row index
        :type row: int
        :param mean: the mean to set
        :type mean: float
        """
        javabridge.call(self.jobject, "setMean", "(IID)V", col, row, mean)

    def get_stdev(self, col, row):
        """
        Returns the standard deviation at this location (if valid location).

        :param col: the 0-based column index
        :type col: int
        :param row: the 0-based row index
        :type row: int
        :return: the standard deviation
        :rtype: float
        """
        return javabridge.call(self.jobject, "getStdDev", "(II)D", col, row)

    def set_stdev(self, col, row, stdev):
        """
        Sets the standard deviation at this location (if valid location).

        :param col: the 0-based column index
        :type col: int
        :param row: the 0-based row index
        :type row: int
        :param stdev: the standard deviation to set
        :type stdev: float
        """
        javabridge.call(self.jobject, "setStdDev", "(IID)V", col, row, stdev)

    def average(self, col):
        """
        Returns the average mean at this location (if valid location).

        :param col: the 0-based column index
        :type col: int
        :return: the mean
        :rtype: float
        """
        return javabridge.call(self.jobject, "getAverage", "(I)D", col)

    def to_string_matrix(self):
        """
        Returns the matrix as a string.

        :return: the generated output
        :rtype: str
        """
        return javabridge.call(self.jobject, "toStringMatrix", "()V")

    def to_string_key(self):
        """
        Returns a key for all the col names, for better readability if the names got cut off.

        :return: the key
        :rtype: str
        """
        return javabridge.call(self.jobject, "toStringKey", "()V")

    def to_string_header(self):
        """
        Returns the header of the matrix as a string.

        :return: the header
        :rtype: str
        """
        return javabridge.call(self.jobject, "toStringHeader", "()V")

    def to_string_summary(self):
        """
        returns the summary as string.

        :return: the summary
        :rtype: str
        """
        return javabridge.call(self.jobject, "toStringSummary", "()V")

    def to_string_ranking(self):
        """
        Returns the ranking in a string representation.

        :return: the ranking
        :rtype: str
        """
        return javabridge.call(self.jobject, "toStringRanking", "()V")


class Tester(OptionHandler):
    """
    For generating statistical results from an experiment.
    """

    def __init__(self, classname="weka.experiment.PairedCorrectedTTester", jobject=None, options=None):
        """
        Initializes the specified tester using either the classname or the supplied JB_Object.

        :param classname: the classname of the tester
        :type classname: str
        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        :param options: the list of commandline options to set
        :type options: list
        """
        if jobject is None:
            jobject = Tester.new_instance(classname)
        self.enforce_type(jobject, "weka.experiment.Tester")
        self.columns_determined = False
        self._dataset_columns = ["Key_Dataset"]
        self._run_column = "Key_Run"
        self._fold_column = "Key_Fold"
        self._result_columns = ["Key_Scheme", "Key_Scheme_options", "Key_Scheme_version_ID"]
        super(Tester, self).__init__(jobject=jobject, options=options)

    @property
    def resultmatrix(self):
        """
        Returns the ResultMatrix instance in use.

        :return: the matrix in use
        :rtype: ResultMatrix
        """
        return ResultMatrix(
            jobject=javabridge.call(self.jobject, "getResultMatrix", "()Lweka/experiment/ResultMatrix;"))

    @resultmatrix.setter
    def resultmatrix(self, matrix):
        """
        Sets the ResultMatrix to use.

        :param matrix: the ResultMatrix instance to use
        :type matrix: ResultMatrix
        """
        javabridge.call(self.jobject, "setResultMatrix", "(Lweka/experiment/ResultMatrix;)V", matrix.jobject)

    @property
    def instances(self):
        """
        Returns the data used in the analysis.

        :return: the data in use
        :rtype: Instances
        """
        inst = javabridge.call(self.jobject, "getInstances", "()Lweka/core/Instances;")
        if inst is None:
            return None
        else:
            return Instances(inst)

    @instances.setter
    def instances(self, data):
        """
        Sets the data to use for analysis.

        :param data: the Instances to analyze
        :type data: Instances
        """
        javabridge.call(self.jobject, "setInstances", "(Lweka/core/Instances;)V", data.jobject)
        self.columns_determined = False

    @property
    def dataset_columns(self):
        """
        Returns the list of column names that identify uniquely a dataset.

        :return: the list of attributes names
        :rtype: list
        """
        return self._dataset_columns

    @dataset_columns.setter
    def dataset_columns(self, col_names):
        """
        Sets the list of column names that identify uniquely a dataset.

        :param col_names: the list of attribute names
        :type col_names: list
        """
        self._dataset_columns = col_names[:]

    @property
    def run_column(self):
        """
        Returns the column name that holds the Run number.

        :return: the attribute name
        :rtype: str
        """
        return self._run_column

    @run_column.setter
    def run_column(self, col_name):
        """
        Sets the column name that holds the Run number.

        :param col_name: the attribute name
        :type col_name: str
        """
        self._run_column = col_name

    @property
    def fold_column(self):
        """
        Returns the column name that holds the Fold number.

        :return: the attribute name
        :rtype: str
        """
        return self._fold_column

    @fold_column.setter
    def fold_column(self, col_name):
        """
        Sets the column name that holds the Fold number.

        :param col_name: the attribute name
        :type col_name: str
        """
        self._fold_column = col_name

    @property
    def result_columns(self):
        """
        Returns the list of column names that identify uniquely a result (eg classifier + options + ID).

        :return: the list of attribute names
        :rtype: list
        """
        return self._result_columns

    @result_columns.setter
    def result_columns(self, col_names):
        """
        Sets the list of column names that identify uniquely a result (eg classifier + options + ID).

        :param col_names: the list of attribute names
        :type col_names: list
        """
        self._result_columns = col_names[:]

    def init_columns(self):
        """
        Sets the column indices based on the supplied names if necessary.
        """
        if self.columns_determined:
            return
        data = self.instances
        if data is None:
            print("No instances set, cannot determine columns!")
            return

        # dataset
        if self.dataset_columns is None:
            raise Exception("No dataset columns set!")
        cols = ""
        for name in self.dataset_columns:
            att = data.attribute_by_name(name)
            if att is None:
                raise Exception("Dataset column not found: " + name)
            if len(cols) > 0:
                cols += ","
            cols += str(att.index + 1)
        javabridge.call(
            self.jobject, "setDatasetKeyColumns", "(Lweka/core/Range;)V", Range(ranges=cols).jobject)

        # run
        if self.run_column is None:
            raise Exception("No run columnn set!")
        att = data.attribute_by_name(self.run_column)
        if att is None:
            raise Exception("Run column not found: " + self.run_column)
        javabridge.call(
            self.jobject, "setRunColumn", "(I)V", att.index)

        # fold
        if not self.fold_column is None:
            att = data.attribute_by_name(self.fold_column)
            if att is None:
                index = -1
            else:
                index = att.index
            javabridge.call(
                self.jobject, "setFoldColumn", "(I)V", index)

        # result
        if self._result_columns is None:
            raise Exception("No reset columns set!")
        cols = ""
        for name in self._result_columns:
            att = data.attribute_by_name(name)
            if att is None:
                raise Exception("Result column not found: " + name)
            if len(cols) > 0:
                cols += ","
            cols += str(att.index + 1)
        javabridge.call(
            self.jobject, "setResultsetKeyColumns", "(Lweka/core/Range;)V", Range(ranges=cols).jobject)

        self.columns_determined = True

    def header(self, comparison_column):
        """
        Creates a "header" string describing the current resultsets.

        :param comparison_column: the index of the column to compare against
        :type comparison_column: int
        :return: the header
        :rtype: str
        """
        self.init_columns()
        return javabridge.call(
            self.jobject, "header", "(I)Ljava/lang/String;", comparison_column)

    def multi_resultset_full(self, base_resultset, comparison_column):
        """
        Creates a comparison table where a base resultset is compared to the other resultsets.

        :param base_resultset: the 0-based index of the base resultset (eg classifier to compare against)
        :type base_resultset: int
        :param comparison_column: the 0-based index of the column to compare against
        :type comparison_column: int
        :return: the comparison
        :rtype: str
        """
        return javabridge.call(
            self.jobject, "multiResultsetFull", "(II)Ljava/lang/String;", base_resultset, comparison_column)

    def multi_resultset_ranking(self, comparison_column):
        """
        Creates a ranking.

        :param comparison_column: the 0-based index of the column to compare against
        :type comparison_column: int
        :return: the ranking
        :rtype: str
        """
        return javabridge.call(
            self.jobject, "multiResultsetRanking", "(I)Ljava/lang/String;", comparison_column)

    def multi_resultset_summary(self, comparison_column):
        """
        Carries out a comparison between all resultsets, counting the number of datsets where one resultset
        outperforms the other.

        :param comparison_column: the 0-based index of the column to compare against
        :type comparison_column: int
        :return: the summary
        :rtype: str
        """
        return javabridge.call(
            self.jobject, "multiResultsetSummary", "(I)Ljava/lang/String;", comparison_column)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# source.py
# Copyright (C) 2015-2016 Fracpete (pythonwekawrapper at gmail dot com)


import os
import re
from weka.core.classes import to_commandline, from_commandline
import weka.flow.base as base
from weka.flow.base import Actor, OutputProducer, Token
from weka.core.database import InstanceQuery
import weka.datagenerators as datagen


class Source(OutputProducer, Actor):
    """
    The ancestor for all sources.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the source.

        :param name: the name of the source
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(Source, self).__init__(name=name, config=config)
        super(OutputProducer, self).__init__(name=name, config=config)


class Start(Source):
    """
    Outputs a None token for triggering other actors.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the source.

        :param name: the name of the source
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(Start, self).__init__(name=name, config=config)

    def description(self):
        """
        Returns a description of the actor.

        :return: the description
        :rtype: str
        """
        return "Outputs a None token for triggering other actors."

    def do_execute(self):
        """
        The actual execution of the actor.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        self._output.append(Token(None))
        return None


class FileSupplier(Source):
    """
    Outputs a fixed list of files.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the source.

        :param name: the name of the source
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(FileSupplier, self).__init__(name=name, config=config)

    def description(self):
        """
        Returns a description of the actor.

        :return: the description
        :rtype: str
        """
        return "Outputs a fixed list of files."

    @property
    def quickinfo(self):
        """
        Returns a short string describing some of the options of the actor.

        :return: the info, None if not available
        :rtype: str
        """
        return "files: " + str(len(self.config["files"]))

    def fix_config(self, options):
        """
        Fixes the options, if necessary. I.e., it adds all required elements to the dictionary.

        :param options: the options to fix
        :type options: dict
        :return: the (potentially) fixed options
        :rtype: dict
        """
        options = super(FileSupplier, self).fix_config(options)

        opt = "files"
        if opt not in options:
            options[opt] = []
        if opt not in self.help:
            self.help[opt] = "The files to output (list of string)."

        return options

    def do_execute(self):
        """
        The actual execution of the actor.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        for f in self.resolve_option("files"):
            self._output.append(Token(f))
        return None


class ListFiles(Source):
    """
    Source that list files in a directory.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the source.

        :param name: the name of the source
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(ListFiles, self).__init__(name=name, config=config)

    def description(self):
        """
        Returns a description of the actor.

        :return: the description
        :rtype: str
        """
        return "Source that list files in a directory."

    @property
    def quickinfo(self):
        """
        Returns a short string describing some of the options of the actor.

        :return: the info, None if not available
        :rtype: str
        """
        return "dir: " + str(self.config["dir"]) \
               + ", files: " + str(self.config["list_files"]) \
               + ", dirs: " + str(self.resolve_option("list_dirs")) \
               + ", recursive: " + str(self.config["recursive"])

    def fix_config(self, options):
        """
        Fixes the options, if necessary. I.e., it adds all required elements to the dictionary.

        :param options: the options to fix
        :type options: dict
        :return: the (potentially) fixed options
        :rtype: dict
        """
        options = super(ListFiles, self).fix_config(options)

        opt = "dir"
        if opt not in options:
            options[opt] = "."
        if opt not in self.help:
            self.help[opt] = "The directory to search (string)."

        opt = "recursive"
        if opt not in options:
            options[opt] = False
        if opt not in self.help:
            self.help[opt] = "Whether to search recursively (bool)."

        opt = "list_files"
        if opt not in options:
            options[opt] = True
        if opt not in self.help:
            self.help[opt] = "Whether to include files (bool)."

        opt = "list_dirs"
        if opt not in options:
            options[opt] = False
        if opt not in self.help:
            self.help[opt] = "Whether to include directories (bool)."

        opt = "regexp"
        if opt not in options:
            options[opt] = ".*"
        if opt not in self.help:
            self.help[opt] = "The regular expression that files/dirs must match (string)."

        return options

    def _list(self, path, collected):
        """
        Lists all the files/dirs in directory that match the pattern.

        :param path: the directory to search
        :type path: str
        :param collected: the files/dirs collected so far (full path)
        :type collected: list
        :return: None if successful, error otherwise
        :rtype: str
        """
        list_files = self.resolve_option("list_files")
        list_dirs = self.resolve_option("list_dirs")
        recursive = self.resolve_option("recursive")
        spattern = str(self.resolve_option("regexp"))
        pattern = None
        if (spattern is not None) and (spattern != ".*"):
            pattern = re.compile(spattern)

        try:
            items = os.listdir(path)
            for item in items:
                fp = path + os.sep + item
                if list_files and os.path.isfile(fp):
                    if (pattern is None) or pattern.match(item):
                        collected.append(fp)
                if list_dirs and os.path.isdir(fp):
                    if (pattern is None) or pattern.match(item):
                        collected.append(fp)
                if recursive and os.path.isdir(fp):
                    self._list(fp, collected)
        except Exception as e:
            return "Error listing '" + path + "': " + str(e)

    def do_execute(self):
        """
        The actual execution of the actor.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        directory = str(self.resolve_option("dir"))
        if not os.path.exists(directory):
            return "Directory '" + directory + "' does not exist!"
        if not os.path.isdir(directory):
            return "Location '" + directory + "' is not a directory!"
        collected = []
        result = self._list(directory, collected)
        if result is None:
            for c in collected:
                self._output.append(Token(c))
        return result


class GetStorageValue(Source):
    """
    Outputs the specified value from storage.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the source.

        :param name: the name of the source
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(GetStorageValue, self).__init__(name=name, config=config)

    def description(self):
        """
        Returns a description of the actor.

        :return: the description
        :rtype: str
        """
        return "Outputs the specified value from storage."

    @property
    def quickinfo(self):
        """
        Returns a short string describing some of the options of the actor.

        :return: the info, None if not available
        :rtype: str
        """
        return "name: " + str(self.config["storage_name"])

    def fix_config(self, options):
        """
        Fixes the options, if necessary. I.e., it adds all required elements to the dictionary.

        :param options: the options to fix
        :type options: dict
        :return: the (potentially) fixed options
        :rtype: dict
        """
        options = super(GetStorageValue, self).fix_config(options)

        opt = "storage_name"
        if opt not in options:
            options[opt] = "unknown"
        if opt not in self.help:
            self.help[opt] = "The name of the storage value to retrieve (string)."

        return options

    def do_execute(self):
        """
        The actual execution of the actor.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        if self.storagehandler is None:
            return "No storage handler available!"
        sname = str(self.resolve_option("storage_name"))
        if sname not in self.storagehandler.storage:
            return "No storage item called '" + sname + "' present!"
        self._output.append(Token(self.storagehandler.storage[sname]))
        return None


class ForLoop(Source):
    """
    Outputs integers using the specified min, max and step.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the source.

        :param name: the name of the source
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(ForLoop, self).__init__(name=name, config=config)

    def description(self):
        """
        Returns a description of the actor.

        :return: the description
        :rtype: str
        """
        return "Outputs integers using the specified min, max and step."

    @property
    def quickinfo(self):
        """
        Returns a short string describing some of the options of the actor.

        :return: the info, None if not available
        :rtype: str
        """
        return "min: " + str(self.config["min"]) \
               + ", max: " + str(self.config["max"]) \
               + ", step: " + str(self.config["step"])

    def fix_config(self, options):
        """
        Fixes the options, if necessary. I.e., it adds all required elements to the dictionary.

        :param options: the options to fix
        :type options: dict
        :return: the (potentially) fixed options
        :rtype: dict
        """
        options = super(ForLoop, self).fix_config(options)

        opt = "min"
        if opt not in options:
            options[opt] = 1
        if opt not in self.help:
            self.help[opt] = "The minimum for the loop (included, int)."

        opt = "max"
        if opt not in options:
            options[opt] = 10
        if opt not in self.help:
            self.help[opt] = "The maximum for the loop (included, int)."

        opt = "step"
        if opt not in options:
            options[opt] = 1
        if opt not in self.help:
            self.help[opt] = "The step size (int)."

        return options

    def do_execute(self):
        """
        The actual execution of the actor.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        for i in range(
                int(self.resolve_option("min")),
                int(self.resolve_option("max")) + 1,
                int(self.resolve_option("step"))):
            self._output.append(Token(i))
        return None


class LoadDatabase(Source):
    """
    Loads a dataset from the database.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the source.

        :param name: the name of the source
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(LoadDatabase, self).__init__(name=name, config=config)
        self._loader = None
        self._iterator = None

    def description(self):
        """
        Returns a description of the actor.

        :return: the description
        :rtype: str
        """
        return "Loads a dataset from a database using a supplied SQL query."

    @property
    def quickinfo(self):
        """
        Returns a short string describing some of the options of the actor.

        :return: the info, None if not available
        :rtype: str
        """
        return "url: " + str(self.config["db_url"]) \
               + ", query: " + str(self.config["query"])

    def fix_config(self, options):
        """
        Fixes the options, if necessary. I.e., it adds all required elements to the dictionary.

        :param options: the options to fix
        :type options: dict
        :return: the (potentially) fixed options
        :rtype: dict
        """
        opt = "db_url"
        if opt not in options:
            options[opt] = "jdbc:mysql://somehost:3306/somedatabase"
        if opt not in self.help:
            self.help[opt] = "The JDBC database URL to connect to (str)."

        opt = "user"
        if opt not in options:
            options[opt] = "user"
        if opt not in self.help:
            self.help[opt] = "The database user to use for connecting (str)."

        opt = "password"
        if opt not in options:
            options[opt] = "secret"
        if opt not in self.help:
            self.help[opt] = "The password for the database user (str)."

        opt = "query"
        if opt not in options:
            options[opt] = "SELECT * FROM table"
        if opt not in self.help:
            self.help[opt] = "The SQL query for generating the dataset (str)."

        opt = "sparse"
        if opt not in options:
            options[opt] = False
        if opt not in self.help:
            self.help[opt] = "Whether to return the data in sparse format (bool)."

        opt = "custom_props"
        if opt not in options:
            options[opt] = ""
        if opt not in self.help:
            self.help[opt] = "Custom properties filename (str)."

        return super(LoadDatabase, self).fix_config(options)

    def do_execute(self):
        """
        The actual execution of the actor.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        iquery = InstanceQuery()
        iquery.db_url = str(self.resolve_option("db_url"))
        iquery.user = str(self.resolve_option("user"))
        iquery.password = str(self.resolve_option("password"))
        props = str(self.resolve_option("custom_props"))
        if (len(props) > 0) and os.path.isfile(props):
            iquery.custom_properties = props
        iquery.query = str(self.resolve_option("query"))
        data = iquery.retrieve_instances()
        self._output.append(Token(data))
        return None


class DataGenerator(Source):
    """
    Generates artificial data using the specified data generator.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the source.

        :param name: the name of the source
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(DataGenerator, self).__init__(name=name, config=config)

    def description(self):
        """
        Returns a description of the actor.

        :return: the description
        :rtype: str
        """
        return "Generates artificial data using the specified data generator."

    @property
    def quickinfo(self):
        """
        Returns a short string describing some of the options of the actor.

        :return: the info, None if not available
        :rtype: str
        """
        return "setup: " + base.to_commandline(self.config["setup"])

    def fix_config(self, options):
        """
        Fixes the options, if necessary. I.e., it adds all required elements to the dictionary.

        :param options: the options to fix
        :type options: dict
        :return: the (potentially) fixed options
        :rtype: dict
        """
        opt = "setup"
        if opt not in options:
            options[opt] = datagen.DataGenerator(classname="weka.datagenerators.classifiers.classification.Agrawal")
        if opt not in self.help:
            self.help[opt] = "The data generator to use (DataGenerator)."

        opt = "incremental"
        if opt not in options:
            options[opt] = False
        if opt not in self.help:
            self.help[opt] = "Whether to output the data incrementally, in case the generator supports that (bool)."

        return super(DataGenerator, self).fix_config(options)

    def to_config(self, k, v):
        """
        Hook method that allows conversion of individual options.

        :param k: the key of the option
        :type k: str
        :param v: the value
        :type v: object
        :return: the potentially processed value
        :rtype: object
        """
        if k == "setup":
            return base.to_commandline(v)
        return super(DataGenerator, self).to_config(k, v)

    def from_config(self, k, v):
        """
        Hook method that allows converting values from the dictionary.

        :param k: the key in the dictionary
        :type k: str
        :param v: the value
        :type v: object
        :return: the potentially parsed value
        :rtype: object
        """
        if k == "setup":
            return from_commandline(v, classname=to_commandline(datagen.DataGenerator()))
        return super(DataGenerator, self).from_config(k, v)

    def do_execute(self):
        """
        The actual execution of the actor.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        generator = datagen.DataGenerator.make_copy(self.resolve_option("setup"))
        generator.dataset_format = generator.define_data_format()
        if bool(self.resolve_option("incremental")) and generator.single_mode_flag:
            for i in range(generator.num_examples_act):
                self._output.append(Token(generator.generate_example()))
        else:
            data = generator.generate_examples()
            self._output.append(Token(data))
        return None


class CombineStorage(Source):
    """
    Expands the storage items specified in format string and forwards the generated string.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the source.

        :param name: the name of the source
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(CombineStorage, self).__init__(name=name, config=config)

    def description(self):
        """
        Returns a description of the actor.

        :return: the description
        :rtype: str
        """
        return "Expands the storage items specified in format string and forwards the generated string."

    @property
    def quickinfo(self):
        """
        Returns a short string describing some of the options of the actor.

        :return: the info, None if not available
        :rtype: str
        """
        return "format: " + str(self.config["format"])

    def fix_config(self, options):
        """
        Fixes the options, if necessary. I.e., it adds all required elements to the dictionary.

        :param options: the options to fix
        :type options: dict
        :return: the (potentially) fixed options
        :rtype: dict
        """
        options = super(CombineStorage, self).fix_config(options)

        opt = "format"
        if opt not in options:
            options[opt] = ""
        if opt not in self.help:
            self.help[opt] = "The format to use for generating the combined string; use '@{blah}' for accessing "\
                             "storage item 'blah' (string)."

        return options

    def do_execute(self):
        """
        The actual execution of the actor.

        :return: None if successful, otherwise error message
        :rtype: str
        """
        formatstr = str(self.resolve_option("format"))
        expanded = self.storagehandler.expand(formatstr)
        self._output.append(Token(expanded))
        return None


class StringConstants(Source):
    """
    Outputs a fixed list of strings.
    """

    def __init__(self, name=None, config=None):
        """
        Initializes the source.

        :param name: the name of the source
        :type name: str
        :param config: the dictionary with the options (str -> object).
        :type config: dict
        """
        super(StringConstants, self).__init__(name=name, config=config)

    def description(self):
        """
        Returns a description of the actor.

        :return: the description
        :rtype: str
        """
        return "Outputs a fixed list of strings."

    @property
    def quickinfo(self):
        """
        Returns a short string describing some of the options of the actor.

        :return: the info, None if not available
        :rtype: str
        """
        return "strings: " + str(len(self.config["strings"]))

    def fix_config(self, options):
        """
        Fixes the options, if necessary. I.e., it adds all required elements to the dictionary.

        :param options: the options to fix
        :type options: dict
        :return: the (potentially) fixed options
        :rtype: dict
        """
        options = super(StringConstants, self).fix_config(options)

        opt = "strings"
        if opt not in options:
            options[opt] = []
        if opt not in self.help:
            self.help[opt] = "The strings to output (list of string)."

        return options

    def do_execute(self):
        """
        The actual execution of the actor.
        
        :return: None if successful, otherwise error message
        :rtype: str
        """
        for s in self.resolve_option("strings"):
            self._output.append(Token(s))
        return None



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# dataset.py
# Copyright (C) 2014-2016 Fracpete (pythonwekawrapper at gmail dot com)

import logging
import weka.plot as plot
if plot.matplotlib_available:
    import matplotlib.pyplot as plt
from weka.core.dataset import Instances

# logging setup
logger = logging.getLogger(__name__)


def scatter_plot(data, index_x, index_y, percent=100.0, seed=1, size=50, title=None, outfile=None, wait=True):
    """
    Plots two attributes against each other.

    TODO: click events http://matplotlib.org/examples/event_handling/data_browser.html

    :param data: the dataset
    :type data: Instances
    :param index_x: the 0-based index of the attribute on the x axis
    :type index_x: int
    :param index_y: the 0-based index of the attribute on the y axis
    :type index_y: int
    :param percent: the percentage of the dataset to use for plotting
    :type percent: float
    :param seed: the seed value to use for subsampling
    :type seed: int
    :param size: the size of the circles in point
    :type size: int
    :param title: an optional title
    :type title: str
    :param outfile: the (optional) file to save the generated plot to. The extension determines the file format.
    :type outfile: str
    :param wait: whether to wait for the user to close the plot
    :type wait: bool
    """
    if not plot.matplotlib_available:
        logger.error("Matplotlib is not installed, plotting unavailable!")
        return

    # create subsample
    data = plot.create_subsample(data, percent=percent, seed=seed)

    # collect data
    x = []
    y = []
    if data.class_index == -1:
        c = None
    else:
        c = []
    for i in range(data.num_instances):
        inst = data.get_instance(i)
        x.append(inst.get_value(index_x))
        y.append(inst.get_value(index_y))
        if c is not None:
            c.append(inst.get_value(inst.class_index))

    # plot data
    fig, ax = plt.subplots()
    if c is None:
        ax.scatter(x, y, s=size, alpha=0.5)
    else:
        ax.scatter(x, y, c=c, s=size, alpha=0.5)
    ax.set_xlabel(data.attribute(index_x).name)
    ax.set_ylabel(data.attribute(index_y).name)
    if title is None:
        title = "Attribute scatter plot"
        if percent != 100:
            title += " (%0.1f%%)" % percent
    ax.set_title(title)
    ax.plot(ax.get_xlim(), ax.get_ylim(), ls="--", c="0.3")
    ax.grid(True)
    fig.canvas.set_window_title(data.relationname)
    plt.draw()
    if outfile is not None:
        plt.savefig(outfile)
    if wait:
        plt.show()


def matrix_plot(data, percent=100.0, seed=1, size=10, title=None, outfile=None, wait=True):
    """
    Plots all attributes against each other.

    TODO: click events http://matplotlib.org/examples/event_handling/data_browser.html

    :param data: the dataset
    :type data: Instances
    :param percent: the percentage of the dataset to use for plotting
    :type percent: float
    :param seed: the seed value to use for subsampling
    :type seed: int
    :param size: the size of the circles in point
    :type size: int
    :param title: an optional title
    :type title: str
    :param outfile: the (optional) file to save the generated plot to. The extension determines the file format.
    :type outfile: str
    :param wait: whether to wait for the user to close the plot
    :type wait: bool
    """
    if not plot.matplotlib_available:
        logger.error("Matplotlib is not installed, plotting unavailable!")
        return

    # create subsample
    data = plot.create_subsample(data, percent=percent, seed=seed)

    fig = plt.figure()

    if data.class_index == -1:
        c = None
    else:
        c = []
        for i in range(data.num_instances):
            inst = data.get_instance(i)
            c.append(inst.get_value(inst.class_index))

    for index_x in range(data.num_attributes):
        x = []
        for i in range(data.num_instances):
            inst = data.get_instance(i)
            x.append(inst.get_value(index_x))
        for index_y in range(data.num_attributes):
            y = []
            for i in range(data.num_instances):
                inst = data.get_instance(i)
                y.append(inst.get_value(index_y))
            ax = fig.add_subplot(
                data.num_attributes, data.num_attributes, index_x * data.num_attributes + index_y + 1)
            if c is None:
                ax.scatter(x, y, s=size, alpha=0.5)
            else:
                ax.scatter(x, y, c=c, s=size, alpha=0.5)
            ax.set_xlabel(data.attribute(index_x).name)
            ax.set_ylabel(data.attribute(index_y).name)
            ax.get_yaxis().set_ticklabels([])
            ax.get_xaxis().set_ticklabels([])
            ax.plot(ax.get_xlim(), ax.get_ylim(), ls="--", c="0.3")
            ax.grid(True)
    if title is None:
        title = data.relationname
    if percent != 100:
        title += " (%0.1f%%)" % percent
    fig.canvas.set_window_title(title)
    plt.draw()
    if outfile is not None:
        plt.savefig(outfile)
    if wait:
        plt.show()


def line_plot(data, atts=None, percent=100.0, seed=1, title=None, outfile=None, wait=True):
    """
    Uses the internal format to plot the dataset, one line per instance.

    :param data: the dataset
    :type data: Instances
    :param atts: the list of 0-based attribute indices of attributes to plot
    :type atts: list
    :param percent: the percentage of the dataset to use for plotting
    :type percent: float
    :param seed: the seed value to use for subsampling
    :type seed: int
    :param title: an optional title
    :type title: str
    :param outfile: the (optional) file to save the generated plot to. The extension determines the file format.
    :type outfile: str
    :param wait: whether to wait for the user to close the plot
    :type wait: bool
    """
    if not plot.matplotlib_available:
        logger.error("Matplotlib is not installed, plotting unavailable!")
        return

    # create subsample
    data = plot.create_subsample(data, percent=percent, seed=seed)

    fig = plt.figure()

    if atts is None:
        x = []
        for i in range(data.num_attributes):
            x.append(i)
    else:
        x = atts

    ax = fig.add_subplot(111)
    ax.set_xlabel("attributes")
    ax.set_ylabel("value")
    ax.grid(True)
    for index_y in range(data.num_instances):
        y = []
        for index_x in x:
            y.append(data.get_instance(index_y).get_value(index_x))
        ax.plot(x, y, "o-", alpha=0.5)
    if title is None:
        title = data.relationname
    if percent != 100:
        title += " (%0.1f%%)" % percent
    fig.canvas.set_window_title(title)
    plt.draw()
    if outfile is not None:
        plt.savefig(outfile)
    if wait:
        plt.show()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# typeconv.py
# Copyright (C) 2014-2021 Fracpete (pythonwekawrapper at gmail dot com)

import javabridge
import logging
import numpy

# logging setup
logger = logging.getLogger(__name__)


def jstring_array_to_list(a):
    """
    Turns the Java string array into Python unicode string list.

    :param a: the string array to convert
    :type a: JB_Object
    :return: the string list
    :rtype: list
    """
    result = []
    length = javabridge.get_env().get_array_length(a)
    wrapped = javabridge.get_env().get_object_array_elements(a)
    for i in range(length):
        result.append(javabridge.get_env().get_string(wrapped[i]))
    return result


def string_list_to_jarray(l):
    """
    Turns a Python unicode string list into a Java String array.

    :param l: the string list
    :type: list
    :rtype: java string array
    :return: JB_Object
    """
    result = javabridge.get_env().make_object_array(len(l), javabridge.get_env().find_class("java/lang/String"))
    for i in range(len(l)):
        javabridge.get_env().set_object_array_element(result, i, javabridge.get_env().new_string_utf(l[i]))
    return result


def jstring_list_to_string_list(l, return_empty_if_none=True):
    """
    Converts a Java java.util.List containing strings into a Python list.

    :param l: the list to convert
    :type l: JB_Object
    :param return_empty_if_none: whether to return an empty list or None when list object is None
    :type return_empty_if_none: bool
    :return: the list with UTF strings
    :rtype: list
    """
    result = []
    if l is None:
        if return_empty_if_none:
            return []
        else:
            return None
    objs = javabridge.get_collection_wrapper(l)
    for obj in objs:
        result.append(javabridge.get_env().get_string_utf(obj))
    return result


def jdouble_matrix_to_ndarray(m):
    """
    Turns the Java matrix (2-dim array) of doubles into a numpy 2-dim array.

    :param m: the double matrix
    :type: JB_Object
    :return: Numpy array
    :rtype: numpy.darray
    """
    rows = javabridge.get_env().get_object_array_elements(m)
    num_rows = len(rows)
    num_cols = javabridge.get_env().get_array_length(rows[0])
    result = numpy.zeros(num_rows * num_cols).reshape((num_rows, num_cols))
    i = 0
    for row in rows:
        elements = javabridge.get_env().get_double_array_elements(row)
        n = 0
        for element in elements:
            result[i][n] = element
            n += 1
        i += 1
    return result


def jenumeration_to_list(enm):
    """
    Turns the java.util.Enumeration into a list.

    :param enm: the enumeration to convert
    :type enm: JB_Object
    :return: the list
    :rtype: list
    """
    result = []
    while javabridge.call(enm, "hasMoreElements", "()Z"):
        result.append(javabridge.call(enm, "nextElement", "()Ljava/lang/Object;"))
    return result


def float_to_jfloat(d):
    """
    Turns the Python float into a Java java.lang.Float object.

    :param d: the Python float
    :type d: float
    :return: the Float object
    :rtype: JB_Object
    """
    return javabridge.make_instance("java/lang/Float", "(D)V", d)


def jdouble_to_float(d):
    """
    Turns the Java java.lang.Double object into Python float object.

    :param d: the java.lang.Double
    :type d: JB_Object
    :return: the Float object
    :rtype: float
    """
    return javabridge.call(d, "doubleValue", "()D", d)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# systeminfo.py
# Copyright (C) 2020 Fracpete (pythonwekawrapper at gmail dot com)

from weka.core.classes import JavaObject
import javabridge


def system_info():
    """
    Returns a dictionary generated from the output of weka.core.SystemInfo.

    :return: the system info as dictionary
    :rtype: dict
    """
    jobj = JavaObject.new_instance("weka.core.SystemInfo")
    info = javabridge.call(jobj, "getSystemInfo", "()Ljava/util/Hashtable;")
    return javabridge.jdictionary_to_string_dictionary(info)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# container.py
# Copyright (C) 2015-2016 Fracpete (pythonwekawrapper at gmail dot com)

import re
from weka.core.dataset import Instances


class Container(object):
    """
    Container for storing multiple objects and passing them around together in the flow.
    """

    def __init__(self):
        """
        Initializes the container.
        """
        self._data = {}
        self._allowed = []

    def get(self, name):
        """
        Returns the stored data.

        :param name: the name of the item to return
        :type name: str
        :return: the data
        :rtype: object
        """
        return self._data[name]

    def set(self, name, value):
        """
        Stores the given data (if not None).

        :param name: the name of the item to store
        :type name: str
        :param value: the value to store
        :type value: object
        """
        if value is not None:
            self._data[name] = value

    @property
    def allowed(self):
        """
        Returns the all the allowed keys.

        :return: the list of allowed keys.
        :rtype: list
        """
        return self._allowed

    def is_valid(self):
        """
        Checks whether the container is valid.

        :return: True if the container is valid
        :rtype: bool
        """
        return True

    def __str__(self):
        """
        Returns the content of the container as string.

        :return: the content
        :rtype: str
        """
        return str(self._data)

    def generate_help(self):
        """
        Generates a help string for this container.

        :return: the help string
        :rtype: str
        """
        result = []
        result.append(self.__class__.__name__)
        result.append(re.sub(r'.', '=', self.__class__.__name__))
        result.append("")
        result.append("Supported value names:")
        for a in self.allowed:
            result.append(a)
        return '\n'.join(result)

    def print_help(self):
        """
        Prints a help string for this actor to stdout.
        """
        print(self.generate_help())


class ModelContainer(Container):
    """
    Container for models.
    """

    def __init__(self, model=None, header=None):
        """
        Initializes the container.

        :param model: the model to store (eg Classifier or Clusterer)
        :type model: object
        :param header: the header instances
        :type header: Instances
        """
        super(ModelContainer, self).__init__()
        self.set("Model", model)
        if header is not None:
            header = Instances.template_instances(header)
        self.set("Header", header)
        self._allowed = ["Model", "Header"]

    def is_valid(self):
        """
        Checks whether the container is valid.

        :return: True if the container is valid
        :rtype: bool
        """
        return ("Model" in self._data) or ("Model" in self._data and "Header" in self._data)


class AttributeSelectionContainer(Container):
    """
    Container for models.
    """

    def __init__(self, original=None, reduced=None, num_atts=None, selected=None, results=None):
        """
        Initializes the container.

        :param original: the original dataset
        :type original: Instances
        :param reduced: the reduced dataset
        :type reduced: Instances
        :param num_atts: the number of attributes
        :type num_atts: int
        :param selected: the list of selected attribute indices (0-based)
        :type selected: list
        :param results: the generated results string
        :type results: str
        """
        super(AttributeSelectionContainer, self).__init__()
        self.set("Original", original)
        self.set("Reduced", reduced)
        self.set("NumAttributes", num_atts)
        self.set("Selected", selected)
        self.set("Results", results)
        self._allowed = ["Original", "Reduced", "NumAttributes", "Selected", "Results"]

    def is_valid(self):
        """
        Checks whether the container is valid.

        :return: True if the container is valid
        :rtype: bool
        """
        return ("Reduced" in self._data) and ("NumAttributes" in self._data) and ("Selected" in self._data)


class ModelContainer(Container):
    """
    Container for models.
    """

    def __init__(self, model=None, header=None):
        """
        Initializes the container.

        :param model: the model to store (eg Classifier or Clusterer)
        :type model: object
        :param header: the header instances
        :type header: Instances
        """
        super(ModelContainer, self).__init__()
        self.set("Model", model)
        if header is not None:
            header = Instances.template_instances(header)
        self.set("Header", header)
        self._allowed = ["Model", "Header"]

    def is_valid(self):
        """
        Checks whether the container is valid.

        :return: True if the container is valid
        :rtype: bool
        """
        return ("Model" in self._data) or ("Model" in self._data and "Header" in self._data)


class ClassificationContainer(Container):
    """
    Container for predictions (classifiers).
    """

    def __init__(self, inst=None, classification=None, label=None, distribution=None):
        """
        Initializes the container.

        :param inst: the instance used for making the prediction
        :type inst: Instance
        :param classification: the classification (numeric value or 0-based label index)
        :type classification: float
        :param label: classification label (for nominal classes)
        :type label: str
        :param distribution: the class distribution
        :type distribution: ndarray
        """
        super(ClassificationContainer, self).__init__()
        self.set("Instance", inst)
        self.set("Classification", classification)
        self.set("Label", label)
        self.set("Distribution", distribution)
        self._allowed = ["Instance", "Classification", "Label", "Distribution"]

    def is_valid(self):
        """
        Checks whether the container is valid.

        :return: True if the container is valid
        :rtype: bool
        """
        return ("Instance" in self._data) and ("Classification" in self._data)


class ClusteringContainer(Container):
    """
    Container for predictions (clusterers).
    """

    def __init__(self, inst=None, cluster=None, distribution=None):
        """
        Initializes the container.

        :param inst: the instance used for making the prediction
        :type inst: Instance
        :param cluster: the cluster
        :type cluster: int
        :param distribution: the class distribution
        :type distribution: ndarray
        """
        super(ClusteringContainer, self).__init__()
        self.set("Instance", inst)
        self.set("Cluster", cluster)
        self.set("Distribution", distribution)
        self._allowed = ["Instance", "Cluster", "Distribution"]

    def is_valid(self):
        """
        Checks whether the container is valid.

        :return: True if the container is valid
        :rtype: bool
        """
        return ("Instance" in self._data) and ("Cluster" in self._data)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# datagenerators.py
# Copyright (C) 2014-2019 Fracpete (pythonwekawrapper at gmail dot com)

import javabridge
import logging
import os
import sys
import argparse
import traceback
import weka.core.jvm as jvm
import weka.core.classes as classes
from weka.core.classes import OptionHandler, join_options, to_commandline, from_commandline
from weka.core.dataset import Instances, Instance

# logging setup
logger = logging.getLogger("weka.datagenerators")


class DataGenerator(OptionHandler):
    """
    Wrapper class for datagenerators.
    """

    def __init__(self, classname="weka.datagenerators.classifiers.classification.Agrawal", jobject=None, options=None):
        """
        Initializes the specified datagenerator using either the classname or the supplied JB_Object.

        :param classname: the classname of the datagenerator
        :type classname: str
        :param jobject: the JB_Object to use
        :type jobject: JB_Object
        """
        if jobject is None:
            jobject = DataGenerator.new_instance(classname)
        self.enforce_type(jobject, "weka.datagenerators.DataGenerator")
        super(DataGenerator, self).__init__(jobject=jobject, options=options)

    def define_data_format(self):
        """
        Returns the data format.

        :return: the data format
        :rtype: Instances
        """
        data = javabridge.call(self.jobject, "defineDataFormat", "()Lweka/core/Instances;")
        if data is None:
            return None
        else:
            return Instances(data)

    @property
    def single_mode_flag(self):
        """
        Returns whether data is generated row by row (True) or in one go (False).

        :return: whether incremental
        :rtype: bool
        """
        return javabridge.call(self.jobject, "getSingleModeFlag", "()Z")

    @property
    def dataset_format(self):
        """
        Returns the dataset format.

        :return: the format
        :rtype: Instances
        """
        data = javabridge.call(self.jobject, "getDatasetFormat", "()Lweka/core/Instances;")
        if data is None:
            return None
        else:
            return Instances(data)

    @dataset_format.setter
    def dataset_format(self, inst):
        """
        Sets the dataset format.

        :param inst: the Instances to use as dataset format
        :type inst: Instances
        """
        javabridge.call(self.jobject, "setDatasetFormat", "(Lweka/core/Instances;)V", inst.jobject)

    def generate_start(self):
        """
        Returns a "start" string.

        :return: the start comment
        :rtype: str
        """
        return javabridge.call(self.jobject, "generateStart", "()Ljava/lang/String;")

    @property
    def num_examples_act(self):
        """
        Returns a actual number of examples to generate.

        :return: the number of examples
        :rtype: int
        """
        return javabridge.call(self.jobject, "getNumExamplesAct", "()I")

    def generate_example(self):
        """
        Returns a single Instance.

        :return: the next example
        :rtype: Instance
        """
        data = javabridge.call(self.jobject, "generateExample", "()Lweka/core/Instance;")
        if data is None:
            return None
        else:
            return Instance(data)

    def generate_examples(self):
        """
        Returns complete dataset.

        :return: the generated dataset
        :rtype: Instances
        """
        data = javabridge.call(self.jobject, "generateExamples", "()Lweka/core/Instances;")
        if data is None:
            return None
        else:
            return Instances(data)

    def generate_finish(self):
        """
        Returns a "finish" string.

        :return: a finish comment
        :rtype: str
        """
        return javabridge.call(self.jobject, "generateFinish", "()Ljava/lang/String;")

    @classmethod
    def make_data(cls, generator, args):
        """
        Generates data using the generator and commandline arguments.

        :param generator: the generator instance to use
        :type generator: DataGenerator
        :param args: the command-line arguments
        :type args: list
        """
        javabridge.static_call(
            "Lweka/datagenerators/DataGenerator;", "makeData",
            "(Lweka/datagenerators/DataGenerator;[Ljava/lang/String;)V",
            generator.jobject, args)

    @classmethod
    def make_copy(cls, generator):
        """
        Creates a copy of the generator.

        :param generator: the generator to copy
        :type generator: DataGenerator
        :return: the copy of the generator
        :rtype: DataGenerator
        """
        return from_commandline(
            to_commandline(generator), classname=classes.get_classname(DataGenerator()))


def main(args=None):
    """
    Runs a datagenerator from the command-line. Calls JVM start/stop automatically.
    Use -h to see all options.

    :param args: the command-line arguments to use, uses sys.argv if None
    :type args: list
    """

    parser = argparse.ArgumentParser(
        description='Executes a data generator from the command-line. Calls JVM start/stop automatically.')
    parser.add_argument("-j", metavar="classpath", dest="classpath", help="additional classpath, jars/directories")
    parser.add_argument("-X", metavar="heap", dest="heap", help="max heap size for jvm, e.g., 512m")
    parser.add_argument("datagenerator", help="data generator classname, e.g., "
                                              + "weka.datagenerators.classifiers.classification.LED24")
    parser.add_argument("option", nargs=argparse.REMAINDER, help="additional data generator options")
    parsed = parser.parse_args(args=args)
    jars = []
    if parsed.classpath is not None:
        jars = parsed.classpath.split(os.pathsep)

    jvm.start(jars, max_heap_size=parsed.heap, packages=True)

    logger.debug("Commandline: " + join_options(sys.argv[1:]))

    try:
        generator = DataGenerator(classname=parsed.datagenerator)
        if len(parsed.option) > 0:
            generator.options = parsed.option
        DataGenerator.make_data(generator, parsed.option)
    except Exception:
        print(traceback.format_exc())
    finally:
        jvm.stop()


def sys_main():
    """
    Runs the main function using the system cli arguments, and
    returns a system error code.

    :return: 0 for success, 1 for failure.
    :rtype: int
    """

    try:
        main()
        return 0
    except Exception:
        print(traceback.format_exc())
        return 1


if __name__ == "__main__":
    try:
        main()
    except Exception:
        print(traceback.format_exc())
